from urllib.error import URLError

import numpy
import smopy
import matplotlib.pyplot as plt
from matplotlib import colors as mcolors
import math
from gtfspy.gtfs import GTFS
from gtfspy.stats import get_spatial_bounds, get_percentile_stop_bounds, get_median_lat_lon_of_stops
from gtfspy.route_types import ROUTE_TYPE_TO_COLOR, ROUTE_TYPE_TO_ZORDER, ROUTE_TYPE_TO_SHORT_DESCRIPTION
import matplotlib as mpl
from matplotlib_scalebar.scalebar import ScaleBar
from gtfspy import util

"""
This module contains functions for plotting (static) visualizations of the public transport networks using matplotlib.
"""
from gtfspy.extended_route_types import ROUTE_TYPE_CONVERSION

MAP_STYLES = [
    "rastertiles/voyager",
    "rastertiles/voyager_nolabels",
    "rastertiles/voyager_only_labels",
    "rastertiles/voyager_labels_under",
    "light_all",
    "dark_all",
    "light_nolabels",
    "light_only_labels",
    "dark_nolabels",
    "dark_only_labels"
]


def _get_median_centered_plot_bounds(g):
    lon_min, lon_max, lat_min, lat_max = get_spatial_bounds(g)
    lat_median, lon_median = get_median_lat_lon_of_stops(g)
    lon_diff = max(abs(lon_median - lon_min), abs(lon_median - lon_max))
    lat_diff = max(abs(lat_median - lat_min), abs(lat_median - lat_max))
    plot_lon_min = lon_median - lon_diff
    plot_lon_max = lon_median + lon_diff
    plot_lat_min = lat_median - lat_diff
    plot_lat_max = lat_median + lat_diff
    return plot_lon_min, plot_lon_max, plot_lat_min, plot_lat_max


def plot_route_network_from_gtfs(g, ax=None, spatial_bounds=None, map_alpha=0.8, scalebar=True, legend=True,
                                 return_smopy_map=False, map_style=None):
    """
    Parameters
    ----------
    g: A gtfspy.gtfs.GTFS object
        Where to get the data from?
    ax: matplotlib.Axes object, optional
        If None, a new figure and an axis is created
    spatial_bounds: dict, optional
        with str keys: lon_min, lon_max, lat_min, lat_max
    return_smopy_map: bool, optional
        defaulting to false

    Returns
    -------
    ax: matplotlib.axes.Axes

    """
    assert(isinstance(g, GTFS))
    route_shapes = g.get_all_route_shapes()

    if spatial_bounds is None:
        spatial_bounds = get_spatial_bounds(g, as_dict=True)
    if ax is not None:
        bbox = ax.get_window_extent().transformed(ax.figure.dpi_scale_trans.inverted())
        width, height = bbox.width, bbox.height
        spatial_bounds = _expand_spatial_bounds_to_fit_axes(spatial_bounds, width, height)
    return plot_as_routes(route_shapes,
                          ax=ax,
                          spatial_bounds=spatial_bounds,
                          map_alpha=map_alpha,
                          plot_scalebar=scalebar,
                          legend=legend,
                          return_smopy_map=return_smopy_map,
                          map_style=map_style)


def plot_as_routes(route_shapes, ax=None, spatial_bounds=None, map_alpha=0.8, plot_scalebar=True, legend=True,
                   return_smopy_map=False, line_width_attribute=None, line_width_scale=1.0, map_style=None):
    """
    Parameters
    ----------
    route_shapes: list of dicts that should have the following keys
            name, type, agency, lats, lons
            with types
            list, list, str, list, list
    ax: axis object
    spatial_bounds: dict
    map_alpha:
    plot_scalebar: bool
    legend:
    return_smopy_map:
    line_width_attribute:
    line_width_scale:

    Returns
    -------
    ax: matplotlib.axes object
    """
    lon_min = spatial_bounds['lon_min']
    lon_max = spatial_bounds['lon_max']
    lat_min = spatial_bounds['lat_min']
    lat_max = spatial_bounds['lat_max']
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)

    smopy_map = get_smopy_map(lon_min, lon_max, lat_min, lat_max, map_style=map_style)
    ax = smopy_map.show_mpl(figsize=None, ax=ax, alpha=map_alpha)
    bound_pixel_xs, bound_pixel_ys = smopy_map.to_pixels(numpy.array([lat_min, lat_max]),
                                                         numpy.array([lon_min, lon_max]))

    route_types_to_lines = {}
    for shape in route_shapes:
        route_type = ROUTE_TYPE_CONVERSION[shape['type']]
        lats = numpy.array(shape['lats'])
        lons = numpy.array(shape['lons'])
        if line_width_attribute:
            line_width = line_width_scale * shape[line_width_attribute]
        else:
            line_width = 1
        xs, ys = smopy_map.to_pixels(lats, lons)
        line, = ax.plot(xs, ys, linewidth=line_width, color=ROUTE_TYPE_TO_COLOR[route_type], zorder=ROUTE_TYPE_TO_ZORDER[route_type])
        route_types_to_lines[route_type] = line

    if legend:
        lines = list(route_types_to_lines.values())
        labels = [ROUTE_TYPE_TO_SHORT_DESCRIPTION[route_type] for route_type in route_types_to_lines.keys()]
        ax.legend(lines, labels, loc="upper left")

    if plot_scalebar:
        _add_scale_bar(ax, lat_max, lon_min, lon_max, bound_pixel_xs.max() - bound_pixel_xs.min())

    ax.set_xticks([])
    ax.set_yticks([])

    ax.set_xlim(bound_pixel_xs.min(), bound_pixel_xs.max())
    ax.set_ylim(bound_pixel_ys.max(), bound_pixel_ys.min())
    if return_smopy_map:
        return ax, smopy_map
    else:
        return ax


def plot_routes_as_stop_to_stop_network(from_lats, from_lons, to_lats, to_lons, attributes=None, color_attributes=None,
                                        zorders=None,
                                        line_labels=None,
                                        ax=None,
                                        spatial_bounds=None,
                                        alpha=1,
                                        map_alpha=0.8,
                                        scalebar=True,
                                        return_smopy_map=False,
                                        c=None, linewidth=None,
                                        linewidth_multiplier=1,
                                        use_log_scale=False):
    if attributes is None:
        attributes = len(list(from_lats))*[None]
        if not linewidth:
            linewidth = 1
    if color_attributes is None:
        color_attributes = len(list(from_lats))*[None]
        assert c is not None
    if zorders is None:
        zorders = len(list(from_lats))*[1]
    if line_labels is None:
        line_labels = len(list(from_lats))*[None]

    if spatial_bounds is None:
        lon_min = min(list(from_lons) + list(to_lons))
        lon_max = max(list(from_lons) + list(to_lons))
        lat_min = min(list(from_lats) + list(to_lats))
        lat_max = max(list(from_lats) + list(to_lats))
    else:
        lon_min = spatial_bounds['lon_min']
        lon_max = spatial_bounds['lon_max']
        lat_min = spatial_bounds['lat_min']
        lat_max = spatial_bounds['lat_max']
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)

    smopy_map = get_smopy_map(lon_min, lon_max, lat_min, lat_max)
    ax = smopy_map.show_mpl(figsize=None, ax=ax, alpha=map_alpha)
    bound_pixel_xs, bound_pixel_ys = smopy_map.to_pixels(numpy.array([lat_min, lat_max]),
                                                         numpy.array([lon_min, lon_max]))

    for from_lat, from_lon, to_lat, to_lon, attribute, color_attribute, zorder, line_label in zip(from_lats,
                                                                                                  from_lons,
                                                                                                  to_lats,
                                                                                                  to_lons,
                                                                                                  attributes,
                                                                                                  color_attributes,
                                                                                                  zorders,
                                                                                                  line_labels):



        if color_attribute is None:
            color = c
        else:
            a = ROUTE_TYPE_CONVERSION[color_attribute]
            color = ROUTE_TYPE_TO_COLOR[a]
            zorder = ROUTE_TYPE_TO_ZORDER[a]
        if not attribute:
            attribute = linewidth
        if use_log_scale:
            attribute = math.log10(attribute)

        xs, ys = smopy_map.to_pixels(numpy.array([from_lat, to_lat]), numpy.array([from_lon, to_lon]))

        ax.plot(xs, ys, color=color, linewidth=attribute*linewidth_multiplier, zorder=zorder, alpha=alpha)
        if line_label:
            ax.text(xs.mean(), ys.mean(), line_label,
                    # verticalalignment='bottom', horizontalalignment='right',
                    color='green', fontsize=15)

    legend = True if color_attributes[0] is not None else False
    import matplotlib.lines as mlines
        
    if legend:
        unique_types = set(color_attributes)
        lines = []
        
        for i in unique_types:
            line = mlines.Line2D([], [], color=ROUTE_TYPE_TO_COLOR[i], markersize=15,
                                 label=ROUTE_TYPE_TO_SHORT_DESCRIPTION[i])

            lines.append(line)
        handles = lines
        labels = [h.get_label() for h in handles] 
            
        ax.legend(handles=handles, labels=labels, loc=4)
        
    if scalebar:
        _add_scale_bar(ax, lat_max, lon_min, lon_max, bound_pixel_xs.max() - bound_pixel_xs.min())

    ax.set_xticks([])
    ax.set_yticks([])

    ax.set_xlim(bound_pixel_xs.min(), bound_pixel_xs.max())
    ax.set_ylim(bound_pixel_ys.max(), bound_pixel_ys.min())

    if return_smopy_map:
        return ax, smopy_map
    else:
        return ax


def _add_scale_bar(ax, lat, lon_min, lon_max, width_pixels):
    distance_m = util.wgs84_distance(lat, lon_min, lat, lon_max)
    scalebar = ScaleBar(distance_m / width_pixels)  # 1 pixel = 0.2 meter
    ax.add_artist(scalebar)


def _expand_spatial_bounds_to_fit_axes(bounds, ax_width, ax_height):
    """
    Parameters
    ----------
    bounds: dict
    ax_width: float
    ax_height: float

    Returns
    -------
    spatial_bounds
    """
    b = bounds
    height_meters = util.wgs84_distance(b['lat_min'], b['lon_min'], b['lat_max'], b['lon_min'])
    width_meters = util.wgs84_distance(b['lat_min'], b['lon_min'], b['lat_min'], b['lon_max'])
    x_per_y_meters = width_meters / height_meters
    x_per_y_axes = ax_width / ax_height
    if x_per_y_axes > x_per_y_meters:  # x-axis
        # axis x_axis has slack -> the spatial longitude bounds need to be extended
        width_meters_new = (height_meters * x_per_y_axes)
        d_lon_new = ((b['lon_max'] - b['lon_min']) / width_meters) * width_meters_new
        mean_lon = (b['lon_min'] + b['lon_max'])/2.
        lon_min = mean_lon - d_lon_new / 2.
        lon_max = mean_lon + d_lon_new / 2.
        spatial_bounds = {
            "lon_min": lon_min,
            "lon_max": lon_max,
            "lat_min": b['lat_min'],
            "lat_max": b['lat_max']
        }
    else:
        # axis y_axis has slack -> the spatial latitude bounds need to be extended
        height_meters_new = (width_meters / x_per_y_axes)
        d_lat_new = ((b['lat_max'] - b['lat_min']) / height_meters) * height_meters_new
        mean_lat = (b['lat_min'] + b['lat_max']) / 2.
        lat_min = mean_lat - d_lat_new / 2.
        lat_max = mean_lat + d_lat_new / 2.
        spatial_bounds = {
            "lon_min": b['lon_min'],
            "lon_max": b['lon_max'],
            "lat_min": lat_min,
            "lat_max": lat_max
        }
    return spatial_bounds



def plot_route_network_thumbnail(g, map_style=None):
    width = 512  # pixels
    height = 300  # pixels
    scale = 24
    dpi = mpl.rcParams["figure.dpi"]

    width_m = width * scale
    height_m = height * scale
    median_lat, median_lon = get_median_lat_lon_of_stops(g)
    dlat = util.wgs84_height(height_m)
    dlon = util.wgs84_width(width_m, median_lat)
    spatial_bounds = {
        "lon_min": median_lon - dlon,
        "lon_max": median_lon + dlon,
        "lat_min": median_lat - dlat,
        "lat_max": median_lat + dlat
    }
    fig = plt.figure(figsize=(width/dpi, height/dpi))
    ax = fig.add_subplot(111)
    plt.subplots_adjust(bottom=0.0, left=0.0, right=1.0, top=1.0)
    return plot_route_network_from_gtfs(g, ax, spatial_bounds, map_alpha=1.0, scalebar=False, legend=False, map_style=map_style)


def plot_stops_with_categorical_attributes(lats_list, lons_list, attributes_list, s=0.5, spatial_bounds=None,  colorbar=False, ax=None, cmap=None, norm=None, alpha=None):
    if not spatial_bounds:
        lon_min = min([min(x) for x in lons_list])
        lon_max = max([max(x) for x in lons_list])
        lat_min = min([min(x) for x in lats_list])
        lat_max = max([max(x) for x in lats_list])
    else:
        lon_min = spatial_bounds['lon_min']
        lon_max = spatial_bounds['lon_max']
        lat_min = spatial_bounds['lat_min']
        lat_max = spatial_bounds['lat_max']
    smopy_map = get_smopy_map(lon_min, lon_max, lat_min, lat_max)
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)
    min_x = max_x = min_y = max_y = None
    for lat in [lat_min, lat_max]:
        for lon in [lon_min, lon_max]:
            x, y = smopy_map.to_pixels(lat, lon)
            if not min_x:
                min_x = x
                max_x = x
                min_y = y
                max_y = y
            else:
                max_x = max(max_x, x)
                max_y = max(max_y, y)
                min_y = min(min_y, y)
                min_x = min(min_x, x)

    ax.set_xlim(min_x, max_x)
    ax.set_ylim(max_y, min_y)
    ax.set_xticks([])
    ax.set_yticks([])
    ax = smopy_map.show_mpl(figsize=None, ax=ax, alpha=0.8)

    axes = []
    for lats, lons, attributes, c in zip(lats_list, lons_list, attributes_list, mcolors.BASE_COLORS):
        x, y = zip(*[smopy_map.to_pixels(lat, lon) for lat, lon in zip(lats, lons)])
        ax = plt.scatter(x, y, s=s, c=c) #, marker=".")
        axes.append(ax)
    return axes


def plot_stops_with_attributes(lats, lons, attribute, s=0.5, spatial_bounds=None, colorbar=False, ax=None, cmap=None, norm=None, alpha=None):
    if not spatial_bounds:
        lon_min = min(lons)
        lon_max = max(lons)
        lat_min = min(lats)
        lat_max = max(lats)
    else:
        lon_min = spatial_bounds['lon_min']
        lon_max = spatial_bounds['lon_max']
        lat_min = spatial_bounds['lat_min']
        lat_max = spatial_bounds['lat_max']
    smopy_map = get_smopy_map(lon_min, lon_max, lat_min, lat_max)
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)
    ax = smopy_map.show_mpl(figsize=None, ax=ax, alpha=0.8)

    xs, ys = smopy_map.to_pixels(lats, lons)
    cax = ax.scatter(xs, ys, c=attribute, s=s, cmap=cmap, norm=norm, alpha=alpha)

    ax.set_xlim(min(xs), max(xs))
    ax.set_ylim(max(ys), min(ys))
    if colorbar:
        return ax, cax, smopy_map
    return ax


def plot_all_stops(g, ax=None, scalebar=False):
    """
    Parameters
    ----------
    g: A gtfspy.gtfs.GTFS object
    ax: matplotlib.Axes object, optional
        If None, a new figure and an axis is created, otherwise results are plotted on the axis.
    scalebar: bool, optional
        Whether to include a scalebar to the plot.

    Returns
    -------
    ax: matplotlib.Axes

    """
    assert(isinstance(g, GTFS))
    lon_min, lon_max, lat_min, lat_max = get_spatial_bounds(g)
    smopy_map = get_smopy_map(lon_min, lon_max, lat_min, lat_max)
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)
    ax = smopy_map.show_mpl(figsize=None, ax=ax, alpha=0.8)

    stops = g.stops()
    lats = numpy.array(stops['lat'])
    lons = numpy.array(stops['lon'])

    xs, ys = smopy_map.to_pixels(lats, lons)
    ax.scatter(xs, ys, color="red", s=10)

    ax.set_xlim(min(xs), max(xs))
    ax.set_ylim(max(ys), min(ys))
    return ax


def get_smopy_map(lon_min, lon_max, lat_min, lat_max, z=None, map_style=None):
    ORIG_TILE_SERVER = smopy.TILE_SERVER
    if map_style is not None:
        assert map_style in MAP_STYLES, map_style + \
                                        " (map_style parameter) is not a valid CartoDB mapping style. Options are " + \
                                        str(MAP_STYLES)
        smopy.TILE_SERVER = "http://1.basemaps.cartocdn.com/" + map_style + "/{z}/{x}/{y}.png"

    args = (lat_min, lat_max, lon_min, lon_max, map_style, z)
    if args not in get_smopy_map.maps:
        kwargs = {}
        if z is not None:  # this hack may not work
            smopy.Map.get_allowed_zoom = lambda self, z: z
            kwargs['z'] = z
        try:
            get_smopy_map.maps[args] = smopy.Map((lat_min, lon_min, lat_max, lon_max), **kwargs)
        except URLError:
            raise RuntimeError("\n Could not load background map from the tile server: "
                               + smopy.TILE_SERVER +
                               "\n Please check that the tile server exists and "
                               "that your are connected to the internet.")
    smopy.TILE_SERVER = ORIG_TILE_SERVER
    return get_smopy_map.maps[args]


get_smopy_map.maps = {}



from gtfspy.import_loaders.table_loader import TableLoader


class DayTripsMaterializer(TableLoader):
    """Make the table day_trips with (date, trip_I, start, end, day_start_ut).

    This replaces the old day_trips view.  This allows direct querying
    on the start_time_ut and end_time_ut, at the cost that this table is
    now O(days * trips).  This makes the following things:

    day_trips2: The actual table
    day_trips: Replacement for the old day_trips view.  day_trips2+trips
    day_stop_times: day_trips2+trips+stop_times
    """
    fname = None
    table = 'day_trips2'
    tabledef = ('(date TEXT, '
                'trip_I INT, '
                'start_time_ut INT, '
                'end_time_ut INT, '
                'day_start_ut INT)')
    copy_where = 'WHERE  {start_ut} < end_time_ut  AND  start_time_ut < {end_ut}'

    @classmethod
    def post_import_round2(cls, conn):
        insert_data_to_day_trips2(conn)

    def index(cls, cur):
        create_day_trips_indices(cur)


    @classmethod
    def make_views(cls, conn):
        """Create day_trips and day_stop_times views.

        day_trips:  day_trips2 x trips  = days x trips
        day_stop_times: day_trips2 x trips x stop_times = days x trips x stop_times
        """
        conn.execute('DROP VIEW IF EXISTS main.day_trips')
        conn.execute('CREATE VIEW day_trips AS   '
                     'SELECT day_trips2.*, trips.* '
                     #'days.day_start_ut+trips.start_time_ds AS start_time_ut, '
                     #'days.day_start_ut+trips.end_time_ds AS end_time_ut   '
                     'FROM day_trips2 JOIN trips USING (trip_I);')
        conn.commit()

        conn.execute('DROP VIEW IF EXISTS main.day_stop_times')
        conn.execute('CREATE VIEW day_stop_times AS   '
                     'SELECT day_trips2.*, trips.*, stop_times.*, '
                     #'days.day_start_ut+trips.start_time_ds AS start_time_ut, '
                     #'days.day_start_ut+trips.end_time_ds AS end_time_ut, '
                     'day_trips2.day_start_ut+stop_times.arr_time_ds AS arr_time_ut, '
                     'day_trips2.day_start_ut+stop_times.dep_time_ds AS dep_time_ut   '
                     'FROM day_trips2 '
                     'JOIN trips USING (trip_I) '
                     'JOIN stop_times USING (trip_I)')
        conn.commit()


def create_day_trips_indices(cur):
    cur.execute('CREATE INDEX IF NOT EXISTS idx_day_trips2_tid '
                'ON day_trips2 (trip_I)')
    cur.execute('CREATE INDEX IF NOT EXISTS idx_day_trips2_d '
            'ON day_trips2 (date)')
    cur.execute('CREATE INDEX IF NOT EXISTS idx_day_trips2_stut_etut '
            'ON day_trips2 (start_time_ut, end_time_ut)')
    # This index may not be needed anymore.
    cur.execute('CREATE INDEX IF NOT EXISTS idx_day_trips2_dsut '
            'ON day_trips2 (day_start_ut)')

def drop_day_trip_indices(cur):
    cur.execute('DROP INDEX IF EXISTS idx_day_trips2_tid')
    cur.execute('DROP INDEX IF EXISTS idx_day_trips2_d')
    cur.execute('DROP INDEX IF EXISTS idx_day_trips2_stut_etut')
    cur.execute('DROP INDEX IF EXISTS idx_day_trips2_dsut')

def insert_data_to_day_trips2(conn):
    cur = conn.cursor()
    cur.execute('DELETE FROM day_trips2')
    cur.execute('INSERT INTO day_trips2 '
                'SELECT date, trip_I, '
                'days.day_start_ut+trips.start_time_ds AS start_time_ut, '
                'days.day_start_ut+trips.end_time_ds AS end_time_ut, '
                'day_start_ut '
                'FROM days JOIN trips USING (trip_I)')
    # Delete rows, where start_time_ut or end_time_ut IS NULL.
    # This could happen e.g. if stop_times are missing for some trip.
    cur.execute("DELETE FROM day_trips2 WHERE start_time_ut IS NULL or end_time_ut IS NULL")
    conn.commit()

def recreate_day_trips2_table(conn):
    drop_day_trip_indices(conn.cursor())
    insert_data_to_day_trips2(conn)
    create_day_trips_indices(conn.cursor())




class Segments(object):

    def __init__(self, gtfs):
        self._gtfs

    def get_segments(self):
        """
        Get segment

        Returns
        -------
        segments: list[Segment]
        """
        cur = self._gtfs.get_cursor()

        # Find our IDs that are relevant.
        cur.execute('''SELECT trip_I, cnt, seq1, seq2, S1.code, S2.code,
                              S1.name AS name1,
                              S2.name AS name2,
                              S1.lat, S1.lon, S2.lat, S2.lon
                       FROM ( SELECT st1.trip_I,  st1.seq AS seq1,  st2.seq AS seq2,
                                  count(*) AS cnt,  st1.arr_time AS at1,
                                  st1.stop_I AS sid1,   st2.stop_I AS sid2
                              FROM calendar LEFT JOIN trips USING (service_I)
                                  JOIN  stop_times st1 ON (trips.trip_I=st1.trip_I)
                                  JOIN stop_times st2 ON (st1.trip_I = st2.trip_I AND st1.seq = st2.seq-1)
                                  LEFT JOIN trips USING (trip_I)
                              WHERE %s==1 AND start_date<=? AND ?<=end_date
                                  AND st1.arr_time_hour=? GROUP BY sid1, sid2 )
                       LEFT JOIN stops S1 ON (sid1=S1.stop_I)
                       LEFT JOIN stops S2 ON (sid2=S2.stop_I)
                       --ORDER BY cnt DESC LIMIT 10 ;
                   ''')



class Segment(object):

    def __init__(self, from_node, to_node, distance, time, vehicle_count, capacity_per_hour, lines, modes):
        self.from_node = from_node
        self.to_node = to_node
        self.distance = distance
        self.time = time
        self.vehicle_count = vehicle_count
        self.capacity_per_hour = capacity_per_hour
        self.lines = lines
        self.modes = modes




from shapely.geometry import MultiPoint


def get_convex_hull_coordinates(gtfs):
    """

    Parameters
    ----------
    gtfs: gtfs.GTFS

    Returns
    -------
    lons: list
        of floats
    lats: list
        of floats
    """
    lons, lats = _get_stop_lat_lons(gtfs)
    lon_lats = list(zip(lons, lats))
    polygon = MultiPoint(lon_lats).convex_hull
    hull_lons, hull_lats= polygon.exterior.coords.xy
    return hull_lats, hull_lons

def _get_stop_lat_lons(gtfs):
    stops = gtfs.stops()
    lats = stops['lat']
    lons = stops['lon']
    return lons, lats

def get_approximate_convex_hull_area_km2(gtfs):
    lons, lats = _get_stop_lat_lons(gtfs)
    return approximate_convex_hull_area(lons, lats)

def approximate_convex_hull_area(lons, lats):
    lon_meters, lat_meters = _get_lon_lat_meters(lons, lats)
    lon_lat_meters = list(zip(lon_meters, lat_meters))
    return MultiPoint(lon_lat_meters).convex_hull.area / 1000 ** 2

def _get_lon_lat_meters(lons, lats):
    lat_min = min(lats)
    lat_max = max(lats)
    lat_mean = (lat_max + lat_min) / 2.
    lon_min = min(lons)
    lon_max = max(lons)
    lon_mean = (lon_max + lon_min) / 2.

    from gtfspy.util import wgs84_distance
    lat_span_meters = wgs84_distance(lat_min, lon_mean, lat_max, lon_mean)
    lon_span_meters = wgs84_distance(lat_mean, lon_min, lat_mean, lon_max)

    lat_meters = [(lat - lat_min) / (lat_max - lat_min) * lat_span_meters for lat in lats]
    lon_meters = [(lon - lon_min) / (lon_max - lon_min) * lon_span_meters for lon in lons]
    return lon_meters, lat_meters


def get_buffered_area_of_stops(gtfs, buffer_meters, resolution):
    """

    Compute the total area of all buffered stops in PT network.

    Parameters
    ----------
    gtfs: gtfs.GTFS
    buffer_meters: meters around the stop to buffer.
    resolution: increases the accuracy of the calculated area with computation time. Default = 16 

    Returns
    -------
    Total area covered by the buffered stops in square meters.
    """
    lons, lats = _get_stop_lat_lons(gtfs)
    a = compute_buffered_area_of_stops(lats, lons, buffer_meters, resolution)
    return a


def compute_buffered_area_of_stops(lats, lons, buffer_meters, resolution=16):
    # geo_series = gp.GeoSeries([Point(lon, lat) for lon, lat in zip(lons, lats)])
    # geo_series.crs = {'init' :'epsg:4326'}
    # geo_series = geo_series.to_crs({'init':'epsg:3857'})

    # circles = geo_series.buffer(buffer_meters, resolution=resolution)
    # multi_points = circles.unary_union
    # return multi_points.area

    if len(lons) > 1:
        lon_meters, lat_meters = _get_lon_lat_meters(lons, lats)
    else:
        lon_meters = lons
        lat_meters = lats

    return MultiPoint(points=list(zip(lon_meters, lat_meters))).buffer(buffer_meters, resolution=resolution).area




# reload(sys)
# -*- encoding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from gtfspy.import_loaders import AgencyLoader, CalendarDatesLoader, CalendarLoader, DayLoader, \
    DayTripsMaterializer, FeedInfoLoader, FrequenciesLoader, TripLoader, MetadataLoader, RouteLoader, \
    ShapeLoader, StopDistancesLoader, StopLoader, StopTimesLoader, TransfersLoader
from gtfspy.import_loaders.table_loader import ignore_tables, decode_six

"""
Importing GTFS into a sqlite database.

Entry point: see main part at the bottom and/or the import_gtfs function.
"""

import re
import sqlite3
import time
from six import string_types

from gtfspy import stats
from gtfspy import util
from gtfspy.gtfs import GTFS


Loaders = [AgencyLoader,  # deps: -
           RouteLoader,  # deps: Agency
           MetadataLoader,  # deps: -
           CalendarLoader,  # deps: -
           CalendarDatesLoader,  # deps: Calendar
           ShapeLoader,  # deps: -
           FeedInfoLoader,  # deps: -
           StopLoader,  # deps: -
           TransfersLoader,  # deps: Stop
           StopDistancesLoader,  # deps: (pi: Stop)
           TripLoader,  # deps: Route, Calendar, (Shape)             | (pi2: StopTimes)
           StopTimesLoader,  # deps: Stop, Trip                           |                  |(v: Trip, Day)
           FrequenciesLoader,  # deps: Trip (pi: Trip, StopTimes)           |
           DayLoader,  # deps: (pi: Calendar, CalendarDates, Trip)  |
           DayTripsMaterializer  # deps:                                      | (pi2: Day)
           ]
postprocessors = [
    #validate_day_start_ut,
]


def import_gtfs(gtfs_sources, output, preserve_connection=False,
                print_progress=True, location_name=None, **kwargs):
    """Import a GTFS database

    gtfs_sources: str, dict, list
        Paths to the gtfs zip file or to the directory containing the GTFS data.
        Alternatively, a dict can be provide that maps gtfs filenames
        (like 'stops.txt' and 'agencies.txt') to their string presentations.

    output: str or sqlite3.Connection
        path to the new database to be created, or an existing
        sqlite3 connection
    preserve_connection: bool, optional
        Whether to close the connection in the end, or not.
    print_progress: bool, optional
        Whether to print progress output
    location_name: str, optional
        set the location of this database
    """
    if isinstance(output, sqlite3.Connection):
        conn = output
    else:
        # if os.path.isfile(output):
        #  raise RuntimeError('File already exists')
        conn = sqlite3.connect(output)
    if not isinstance(gtfs_sources, list):
        gtfs_sources = [gtfs_sources]
    cur = conn.cursor()
    time_import_start = time.time()

    # These are a bit unsafe, but make importing much faster,
    # especially on scratch.
    cur.execute('PRAGMA page_size = 4096;')
    cur.execute('PRAGMA mmap_size = 1073741824;')
    cur.execute('PRAGMA cache_size = -2000000;')
    cur.execute('PRAGMA temp_store=2;')
    # Changes of isolation level are python3.6 workarounds -
    # eventually will probably be fixed and this can be removed.
    conn.isolation_level = None  # change to autocommit mode (former default)
    cur.execute('PRAGMA journal_mode = OFF;')
    #cur.execute('PRAGMA journal_mode = WAL;')
    cur.execute('PRAGMA synchronous = OFF;')
    conn.isolation_level = ''    # change back to python default.
    # end python3.6 workaround

    # Do the actual importing.
    loaders = [L(gtfssource=gtfs_sources, print_progress=print_progress, **kwargs) for L in Loaders]

    for loader in loaders:
        loader.assert_exists_if_required()

    # Do initial import.  This consists of making tables, raw insert
    # of the CSVs, and then indexing.

    for loader in loaders:
        loader.import_(conn)

    # Do any operations that require all tables present.
    for Loader in loaders:
        Loader.post_import_round2(conn)

    # Make any views
    for Loader in loaders:
        Loader.make_views(conn)

    # Make any views
    for F in postprocessors:
        F(conn)

    # Set up same basic metadata.
    from gtfspy import gtfs as mod_gtfs
    G = mod_gtfs.GTFS(output)
    G.meta['gen_time_ut'] = time.time()
    G.meta['gen_time'] = time.ctime()
    G.meta['import_seconds'] = time.time() - time_import_start
    G.meta['download_date'] = ''
    G.meta['location_name'] = ''
    G.meta['n_gtfs_sources'] = len(gtfs_sources)

    # Extract things from GTFS
    download_date_strs = []
    for i, source in enumerate(gtfs_sources):
        if len(gtfs_sources) == 1:
            prefix = ""
        else:
            prefix = "feed_" + str(i) + "_"
        if isinstance(source, string_types):
            G.meta[prefix + 'original_gtfs'] = decode_six(source) if source else None
            # Extract GTFS date.  Last date pattern in filename.
            filename_date_list = re.findall(r'\d{4}-\d{2}-\d{2}', source)
            if filename_date_list:
                date_str = filename_date_list[-1]
                G.meta[prefix + 'download_date'] = date_str
                download_date_strs.append(date_str)
            if location_name:
                G.meta['location_name'] = location_name
            else:
                location_name_list = re.findall(r'/([^/]+)/\d{4}-\d{2}-\d{2}', source)
                if location_name_list:
                    G.meta[prefix + 'location_name'] = location_name_list[-1]
                else:
                    try:
                        G.meta[prefix + 'location_name'] = source.split("/")[-4]
                    except:
                        G.meta[prefix + 'location_name'] = source

    if G.meta['download_date'] == "":
        unique_download_dates = list(set(download_date_strs))
        if len(unique_download_dates) == 1:
            G.meta['download_date'] = unique_download_dates[0]

    G.meta['timezone'] = cur.execute('SELECT timezone FROM agencies LIMIT 1').fetchone()[0]
    stats.update_stats(G)
    del G

    if print_progress:
        print("Vacuuming...")
    # Next 3 lines are python 3.6 work-arounds again.
    conn.isolation_level = None  # former default of autocommit mode
    cur.execute('VACUUM;')
    conn.isolation_level = ''    # back to python default
    # end python3.6 workaround
    if print_progress:
        print("Analyzing...")
    cur.execute('ANALYZE')
    if not (preserve_connection is True):
        conn.close()

def validate_day_start_ut(conn):
    """This validates the day_start_ut of the days table."""
    G = GTFS(conn)
    cur = conn.execute('SELECT date, day_start_ut FROM days')
    for date, day_start_ut in cur:
        #print date, day_start_ut
        assert day_start_ut == G.get_day_start_ut(date)


def main_make_views(gtfs_fname):
    """Re-create all views.
    """
    print("creating views")
    conn = GTFS(fname_or_conn=gtfs_fname).conn
    for L in Loaders:
        L(None).make_views(conn)
    conn.commit()

def main():
    import argparse

    parser = argparse.ArgumentParser(description="""
    Import GTFS files.  Imports gtfs.  There are two subcommands.  The
    'import' subcommand converts from a GTFS directory or zipfile to a
    sqlite database.  Both must be specified on the command line.  The
    import-auto subcommand is older, and uses the logic in code/db.py to
    automatically find databases and output files (in scratch/gtfs and
    scratch/db) based on the shortname given on the command line.  This
    should probably not be used much anymore, instead automate that before
    calling this program.""")
    subparsers = parser.add_subparsers(dest='cmd')
    # parsing import
    parser_import = subparsers.add_parser('import', help="Direct import GTFS->sqlite")
    parser_import.add_argument('gtfs', help='Input GTFS filename (dir or .zip)')
    parser_import.add_argument('output', help='Output .sqlite filename (must end in .sqlite)')
    parser.add_argument('--fast', action='store_true',
                        help='Skip stop_times and shapes tables.')

    # parsing import-auto
    parser_importauto = subparsers.add_parser('import-auto', help="Automatic GTFS import from files")
    parser_importauto.add_argument('gtfsname', help='Input GTFS filename')

    # parsing import-multiple
    parser_import_multiple = subparsers.add_parser('import-multiple', help="GTFS import from multiple zip-files")
    parser_import_multiple.add_argument('zipfiles', metavar='zipfiles', type=str, nargs=argparse.ONE_OR_MORE,
                                        help='zipfiles for the import')
    parser_import_multiple.add_argument('output', help='Output .sqlite filename (must end in .sqlite)')

    # parsing import-list

    # Parsing copy
    parser_copy = subparsers.add_parser('copy', help="Copy database")
    parser_copy.add_argument('source', help='Input GTFS .sqlite')
    parser_copy.add_argument('dest', help='Output GTFS .sqlite')
    parser_copy.add_argument('--start', help='Start copy time (inclusive)')
    parser_copy.add_argument('--end', help='Start copy time (exclusive)')

    # Parsing copy
    parser_copy = subparsers.add_parser('make-views', help="Re-create views")
    parser_copy.add_argument('gtfs', help='Input GTFS .sqlite')

    # make-weekly-download
    parser_copy = subparsers.add_parser('make-weekly')
    parser_copy.add_argument('source', help='Input GTFS .sqlite')
    parser_copy.add_argument('dest', help='Output GTFS .sqlite')

    parser_copy = subparsers.add_parser('make-daily')
    parser_copy.add_argument('source', help='Input GTFS .sqlite')
    parser_copy.add_argument('dest', help='Output GTFS .sqlite')

    # Export stop distances
    parser_copy = subparsers.add_parser('export-stop-distances')
    parser_copy.add_argument('gtfs', help='Input GTFS .sqlite')
    parser_copy.add_argument('output', help='Output for .txt file')

    # Custom stuff
    parser_copy = subparsers.add_parser('custom')
    parser_copy.add_argument('gtfs', help='Input GTFS .sqlite')

    args = parser.parse_args()
    if args.fast:
        ignore_tables.update(('stop_times', 'shapes'))
    # if the first argument is import, import a GTFS directory to a .sqlite database.
    # Both directory and
    if args.cmd == 'import':
        gtfs = args.gtfs
        output = args.output
        # This context manager makes a tmpfile for import.  If there
        # is corruption during import, it won't leave a incomplete or
        # corrupt file where it will be noticed.
        with util.create_file(output, tmpdir=True, keepext=True) as tmpfile:
            import_gtfs(gtfs, output=tmpfile)
    elif args.cmd == "import-multiple":
        zipfiles = args.zipfiles
        output = args.output
        print("loaders")
        with util.create_file(output, tmpdir=True, keepext=True) as tmpfile:
            import_gtfs(zipfiles, output=tmpfile)
    elif args.cmd == 'make-views':
        main_make_views(args.gtfs)
    # This is now implemented in gtfs.py, please remove the commented code
    # if no one has touched this in a while.:
    #
    # elif args.cmd == 'make-weekly':
    #     G = GTFS(args.source)
    #     download_date = G.meta['download_date']
    #     d = datetime.strptime(download_date, '%Y-%m-%d').date()
    #     date_start = d + timedelta(7-d.isoweekday()+1)      # inclusive
    #     date_end   = d + timedelta(7-d.isoweekday()+1 + 7)  # exclusive
    #     G.copy_and_filter(args.dest, start_date=date_start, end_date=date_end)
    # elif args.cmd == 'make-daily':
    #     G = GTFS(args.source)
    #     download_date = G.meta['download_date']
    #     d = datetime.strptime(download_date, '%Y-%m-%d').date()
    #     date_start = d + timedelta(7-d.isoweekday()+1)      # inclusive
    #     date_end   = d + timedelta(7-d.isoweekday()+1 + 1)  # exclusive
    #     G.copy_and_filter(args.dest, start_date=date_start, end_date=date_end)
    elif args.cmd == 'export-stop-distances':
        conn = sqlite3.connect(args.gtfs)
        L = StopDistancesLoader(conn)
        L.export_stop_distances(conn, open(args.output, 'w'))
    elif args.cmd == 'custom':
        pass
        # This is designed for just testing things.  This code should
        # always be commented out in the VCS.
        # conn = sqlite3.connect(args.gtfs)
        # L = StopDistancesLoader(conn)
        # L.post_import(None)
        # L.export_stop_distances(conn, open('sd.txt', 'w'))
    else:
        print("Unrecognized command: %s" % args.cmd)
        exit(1)

if __name__ == "__main__":
    main()



import os
import sqlite3
import unittest
from datetime import datetime

from matplotlib.axes import Axes

from gtfspy.gtfs import GTFS
from gtfspy.import_gtfs import import_gtfs
from gtfspy.plots import plot_trip_counts_per_day


class TestPlots(unittest.TestCase):
    def setUp(self):
        self.gtfs_source_dir = os.path.join(os.path.dirname(__file__), "test_data")
        self.fname = self.gtfs_source_dir + "/test_gtfs.sqlite"
        conn = sqlite3.connect(self.fname)
        import_gtfs(self.gtfs_source_dir, conn, preserve_connection=True, print_progress=False)
        self.G = GTFS(conn)

    def _remove_temporary_files(self):
        for fn in [self.fname]:
            if os.path.exists(fn) and os.path.isfile(fn):
                os.remove(fn)

    def tearDown(self):
        self._remove_temporary_files()

    def test_plot_trip_counts_per_day(self):
        # simple test
        ax = plot_trip_counts_per_day(self.G,
                                      highlight_dates=["2009-01-01"],
                                      highlight_date_labels=["test_date"])
        # test with multiple dates and datetime
        dates = [datetime(2009, month=10, day=1), datetime(2010, month=10, day=1)]
        labels = ["test_date_1", "test_date_2"]
        ax = plot_trip_counts_per_day(self.G,
                                      highlight_dates=dates,
                                      highlight_date_labels=labels)
        assert isinstance(ax, Axes)



from gtfspy.import_loaders.table_loader import TableLoader, decode_six


class CalendarDatesLoader(TableLoader):
    fname = 'calendar_dates.txt'
    table = 'calendar_dates'
    tabledef = '(service_I INTEGER NOT NULL, date TEXT, exception_type INT)'
    copy_where = ("WHERE  date({start_ut}, 'unixepoch', 'localtime') <= date "
                  "AND  date < date({end_ut}, 'unixepoch', 'localtime')")

    def gen_rows(self, readers, prefixes):
        conn = self._conn
        cur = conn.cursor()
        for reader, prefix in zip(readers, prefixes):
            for row in reader:
                date = row['date']
                date_str = '%s-%s-%s' % (date[:4], date[4:6], date[6:8])
                service_id = prefix+row['service_id']
                # We need to find the service_I of this.  To do this we
                # need to check the calendar table, since that (and only
                # that) is the absolute list of service_ids.
                service_I = cur.execute(
                    'SELECT service_I FROM calendar WHERE service_id=?',
                    (decode_six(service_id),)).fetchone()
                if service_I is None:
                    # We have to add a new fake row in order to get a
                    # service_I.  calendar is *the* authoritative source
                    # for service_I:s.
                    cur.execute('INSERT INTO calendar '
                                '(service_id, m,t,w,th,f,s,su, start_date,end_date)'
                                'VALUES (?, 0,0,0,0,0,0,0, ?,?)',
                                (decode_six(service_id), date_str, date_str)
                                )
                    service_I = cur.execute(
                        'SELECT service_I FROM calendar WHERE service_id=?',
                        (decode_six(service_id),)).fetchone()
                service_I = service_I[0]  # row tuple -> int

                yield dict(
                    service_I     = int(service_I),
                    date          = date_str,
                    exception_type= int(row['exception_type']),
                )


from gtfspy.routing.connection import Connection
from gtfspy.networks import temporal_network, walk_transfer_stop_to_stop_network
from gtfspy.gtfs import GTFS
import pandas
from warnings import warn


def get_transit_connections(gtfs, start_time_ut, end_time_ut):
    """
    Parameters
    ----------
    gtfs: gtfspy.GTFS
    end_time_ut: int
    start_time_ut: int

    Returns
    -------
    list[Connection]
    """
    if start_time_ut + 20 * 3600 < end_time_ut:
        warn("Note that it is possible that same trip_I's can take place during multiple days, "
             "which could (potentially) affect the outcomes of the CSA routing!")
    assert (isinstance(gtfs, GTFS))
    events_df = temporal_network(gtfs, start_time_ut=start_time_ut, end_time_ut=end_time_ut)
    assert (isinstance(events_df, pandas.DataFrame))
    return list(map(lambda e: Connection(e.from_stop_I, e.to_stop_I, e.dep_time_ut, e.arr_time_ut, e.trip_I, e.seq),
                    events_df.itertuples()
                    )
                )


def get_walk_network(gtfs, max_link_distance_m=1000):
    """
    Parameters
    ----------
    gtfs: gtfspy.GTFS

    Returns
    -------
    walk_network: networkx.Graph:
    """
    assert (isinstance(gtfs, GTFS))
    return walk_transfer_stop_to_stop_network(gtfs, max_link_distance=max_link_distance_m)



class SpreadingStop:

    def __init__(self, stop_I, min_transfer_time):
        self.stop_I = stop_I
        self.min_transfer_time = min_transfer_time
        self.visit_events = []

    def get_min_visit_time(self):
        """
        Get the earliest visit time of the stop.
        """
        if not self.visit_events:
            return float('inf')
        else:
            return min(self.visit_events, key=lambda event: event.arr_time_ut).arr_time_ut

    def get_min_event(self):
        if len(self.visit_events) == 0:
            return None
        else:
            return min(self.visit_events, key=lambda event: event.arr_time_ut)

    def visit(self, event):
        """
        Visit the stop if it has not been visited already by an event with
        earlier arr_time_ut (or with other trip that does not require a transfer)

        Parameters
        ----------
        event : Event
            an instance of the Event (namedtuple)

        Returns
        -------
        visited : bool
            if visit is stored, returns True, otherwise False
        """
        to_visit = False
        if event.arr_time_ut <= self.min_transfer_time+self.get_min_visit_time():
            to_visit = True
        else:
            for ve in self.visit_events:
                if (event.trip_I == ve.trip_I) and event.arr_time_ut < ve.arr_time_ut:
                    to_visit = True

        if to_visit:
            self.visit_events.append(event)
            min_time = self.get_min_visit_time()
            # remove any visits that are 'too old'
            self.visit_events = [v for v in self.visit_events if v.arr_time_ut <= min_time+self.min_transfer_time]
        return to_visit

    def has_been_visited(self):
        return len(self.visit_events) > 0

    def can_infect(self, event):
        """
        Whether the spreading stop can infect using this event.
        """
        if event.from_stop_I != self.stop_I:
            return False

        if not self.has_been_visited():
            return False
        else:
            time_sep = event.dep_time_ut-self.get_min_visit_time()
            # if the gap between the earliest visit_time and current time is
            # smaller than the min. transfer time, the stop can pass the spreading
            # forward
            if (time_sep >= self.min_transfer_time) or (event.trip_I == -1 and time_sep >= 0):
                return True
            else:
                for visit in self.visit_events:
                    # if no transfer, please hop-on
                    if (event.trip_I == visit.trip_I) and (time_sep >= 0):
                        return True
            return False



import os
import time
from datetime import datetime

from gtfspy.import_loaders.table_loader import TableLoader, decode_six
from gtfspy.util import set_process_timezone


class AgencyLoader(TableLoader):
    fname = 'agency.txt'
    table = 'agencies'
    tabledef = ('(agency_I INTEGER PRIMARY KEY, agency_id TEXT UNIQUE NOT NULL, '
                'name TEXT, url TEXT, timezone TEXT, lang TEXT, phone TEXT)')

    # shape_id,shape_pt_lat,shape_pt_lon,shape_pt_sequence
    # 1001_20140811_1,60.167430,24.951684,1
    def gen_rows(self, readers, prefixes):

        for reader, prefix in zip(readers, prefixes):
            for row in reader:
                yield dict(
                    agency_id     =prefix + decode_six(row.get('agency_id', '1')),
                    name          = decode_six(row['agency_name']),
                    timezone      = decode_six(row['agency_timezone']),
                    url           = decode_six(row['agency_url']),
                    lang          = decode_six(row['agency_lang']) if 'agency_lang' in row else None,
                    phone         = decode_six(row['agency_phone']) if 'agency_phone' in row else None,
                )

    def post_import(self, cur):
        TZs = cur.execute('SELECT DISTINCT timezone FROM agencies').fetchall()
        if len(TZs) == 0:
            raise ValueError("Error: no timezones defined in sources: %s" % self.gtfs_sources)
        elif len(TZs) > 1:
            first_tz = TZs[0][0]
            import pytz
            for tz in TZs[1:]:
                generic_date = datetime(2009, 9, 1)
                ftz = pytz.timezone(first_tz).utcoffset(generic_date, is_dst=True)
                ctz = pytz.timezone(tz[0]).utcoffset(generic_date, is_dst=True)
                if not str(ftz) == str(ctz):
                    raise ValueError("Error: multiple timezones defined in sources:: %s" % self.gtfs_sources)
        TZ = TZs[0][0]
        set_process_timezone(TZ)

    def index(self, cur):
        pass


import os
import unittest
import tempfile as temp

import pandas as pd

from gtfspy.gtfs import GTFS
from gtfspy import stats


class StatsTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """ This method is run once before executing any tests"""
        cls.gtfs_source_dir = os.path.join(os.path.dirname(__file__), "test_data")
        cls.G = GTFS.from_directory_as_inmemory_db(cls.gtfs_source_dir)

    def setUp(self):
        """This method is run once before _each_ test method is executed"""
        self.gtfs = GTFS.from_directory_as_inmemory_db(self.gtfs_source_dir)

    def test_write_stats_as_csv(self):
        testfile = temp.NamedTemporaryFile(mode='w+b')

        stats.write_stats_as_csv(self.gtfs, testfile.name)
        df = pd.read_csv(testfile.name)
        print('len is ' + str(len(df)))
        self.assertEqual(len(df), 1)

        stats.write_stats_as_csv(self.gtfs, testfile.name)
        df = pd.read_csv(testfile.name)
        self.assertEqual(len(df), 2)
        testfile.close()

    def test_get_stats(self):
        d = stats.get_stats(self.gtfs)
        self.assertTrue(isinstance(d, dict))

    def test_calc_and_store_stats(self):
        self.gtfs.meta['stats_calc_at_ut'] = None
        stats.update_stats(self.gtfs)
        self.assertTrue(isinstance(stats.get_stats(self.gtfs), dict))
        self.assertTrue(self.G.meta['stats_calc_at_ut'] is not None)

    def test_get_median_lat_lon_of_stops(self):
        lat, lon = stats.get_median_lat_lon_of_stops(self.gtfs)
        self.assertTrue(lat != lon, "probably median lat and median lon should not be equal for any real data set")
        self.assertTrue(isinstance(lat, float))
        self.assertTrue(isinstance(lon, float))

    def test_get_centroid_of_stops(self):
        lat, lon = stats.get_centroid_of_stops(self.gtfs)
        self.assertTrue(lat != lon, "probably centroid lat and lon should not be equal for any real data set")
        self.assertTrue(isinstance(lat, float))
        self.assertTrue(isinstance(lon, float))

    def test_trip_stats(self):
        df = stats.trip_stats(self.gtfs)
        self.assertTrue(isinstance(df, pd.DataFrame))
        self.assertTrue(len(df.columns.values) > 0)
        self.assertTrue(len(df.index) > 0)
        # applying the function with results_by_mode = True, which should produce a dict
        df = stats.trip_stats(self.gtfs, True)
        self.assertTrue(isinstance(df, dict))
        # assuming test data only has one mode type
        self.assertTrue(len(df.keys()) == 2)

    def test_hourly_frequencies(self):
        # Selectiong a randome time frame
        df = stats.hourly_frequencies(self.gtfs, st=21600, et=45000, route_type=3)
        self.assertTrue(isinstance(df, pd.DataFrame))
        self.assertTrue(isinstance(df, object))
        self.assertTrue(len(df.columns), 4)





import unittest
import os
from gtfspy.gtfs import GTFS
from gtfspy.import_validator import ImportValidator


class TestImportValidator(unittest.TestCase):

    def setUp(self):
        # create validator object using textfiles
        test_feed_dir = os.path.join(os.path.dirname(__file__), "test_data/")
        test_feed_b_dir = os.path.join(test_feed_dir, "feed_b")
        self.gtfs_source_dir = os.path.join(os.path.dirname(__file__), "test_data")
        self.G_txt = GTFS.from_directory_as_inmemory_db([test_feed_dir, test_feed_b_dir])
        self.import_validator = ImportValidator([test_feed_dir, test_feed_b_dir], self.G_txt, verbose=False)

    def test_source_gtfsobj_comparison(self):
        self.import_validator._validate_table_row_counts()

    def test_null_counts_in_gtfsobj(self):
        self.import_validator._validate_no_null_values()

    def test_validate(self):
        warnings_container = self.import_validator.validate_and_get_warnings()
        stop_dist_warning_exists = False
        for warning, count in warnings_container.get_warning_counter().most_common():
            if "stop_distances" in warning:
                stop_dist_warning_exists = True
        assert stop_dist_warning_exists







"""
An abstract class defining some requirements for the algorithms inherited by this class.
"""
from abc import ABCMeta, abstractmethod

import time
from gtfspy.routing.util import timeit


class AbstractRoutingAlgorithm:
    __metaclass__ = ABCMeta

    def __init__(self):
        self._run_time = None
        self._has_run = None

    @abstractmethod
    def _run(self):
        pass

    @timeit
    def run(self):
        if self._has_run:
            raise RuntimeError("Algorithm has already run, please initialize a new algorithm")
        start_time = time.time()
        self._run()
        end_time = time.time()
        self._run_time = end_time - start_time
        self._has_run = True

    def get_run_time(self):
        """
        Returns
        -------
        run_time: float
            running time of the algorithm in seconds
        """
        assert self._has_run
        return self._run_time



from gtfspy.import_loaders.table_loader import TableLoader, decode_six


class TripLoader(TableLoader):
    fname = 'trips.txt'
    table = 'trips'
    # service_I INT NOT NULL
    tabledef = ('(trip_I INTEGER PRIMARY KEY, trip_id TEXT UNIQUE NOT NULL, '
                'route_I INT, service_I INT, direction_id TEXT, shape_id TEXT, '
                'headsign TEXT, '
                'start_time_ds INT, end_time_ds INT)')
    extra_keys = ['route_I', 'service_I' ] #'shape_I']
    extra_values = ['(SELECT route_I FROM routes WHERE route_id=:_route_id )',
                    '(SELECT service_I FROM calendar WHERE service_id=:_service_id )',
                    #'(SELECT shape_I FROM shapes WHERE shape_id=:_shape_id )'
                    ]

    # route_id,service_id,trip_id,trip_headsign,direction_id,shape_id,wheelchair_accessible,bikes_allowed
    # 1001,1001_20150424_20150426_Ke,1001_20150424_Ke_1_0953,"Kapyla",0,1001_20140811_1,1,2
    def gen_rows(self, readers, prefixes):
        #try:
        for reader, prefix in zip(readers, prefixes):
            for row in reader:
                #print row
                    yield dict(
                        _route_id     = prefix + decode_six(row['route_id']),
                        _service_id   = prefix + decode_six(row['service_id']),
                        trip_id       = prefix + decode_six(row['trip_id']),
                        direction_id  = decode_six(row['direction_id']) if row.get('direction_id','') else None,
                        shape_id      = prefix + decode_six(row['shape_id']) if row.get('shape_id','') else None,
                        headsign      = decode_six(row['trip_headsign']) if 'trip_headsign' in row else None,
                        )
        #except:
            #print(row)

    @classmethod
    def index(cls, cur):
        # cur.execute('CREATE INDEX IF NOT EXISTS idx_trips_tid ON trips (trip_id)')
        cur.execute('CREATE INDEX IF NOT EXISTS idx_trips_svid ON trips (service_I)')
        cur.execute('CREATE INDEX IF NOT EXISTS idx_trips_rid ON trips (route_I)')

    def post_import_round2(self, conn):
        update_trip_travel_times_ds(conn)

    # This has now been moved to DayTripsMaterializer, but is left
    # here in case we someday want to make DayTripsMaterializer
    # optional.
    #@classmethod
    #def make_views(cls, conn):
    #    conn.execute('DROP VIEW IF EXISTS main.day_trips')
    #    conn.execute('CREATE VIEW day_trips AS   '
    #                 'SELECT trips.*, days.*, '
    #                 'days.day_start_ut+trips.start_time_ds AS start_time_ut, '
    #                 'days.day_start_ut+trips.end_time_ds AS end_time_ut   '
    #                 'FROM days JOIN trips USING (trip_I);')
    #    conn.commit()


def update_trip_travel_times_ds(conn):
    cur0 = conn.cursor()
    cur = conn.cursor()
    cur0.execute('''SELECT trip_I, min(dep_time), max(arr_time)
                   FROM trips JOIN stop_times USING (trip_I)
                   GROUP BY trip_I''')

    print("updating trips travel times")

    def iter_rows(cur0):
        for row in cur0:
            if row[1]:
                st = row[1].split(':')
                start_time_ds = int(st[0]) * 3600 + int(st[1]) * 60 + int(st[2])
            else:
                start_time_ds = None
            if row[2]:
                et = row[2].split(':')
                end_time_ds = int(et[0]) * 3600 + int(et[1]) * 60 + int(et[2])
            else:
                end_time_ds = None
            yield start_time_ds, end_time_ds, row[0]

    cur.executemany('''UPDATE trips SET start_time_ds=?, end_time_ds=? WHERE trip_I=?''',
                    iter_rows(cur0))
    conn.commit()



import pandas
from matplotlib import pyplot as plt

"""
A collection of various useful plots.
"""


def plot_trip_counts_per_day(G, ax=None, highlight_dates=None, highlight_date_labels=None, show=False):
    """
    Parameters
    ----------
    G: gtfspy.GTFS
    ax: maptlotlib.Axes, optional
    highlight_dates: list[str|datetime.datetime]
        The values of highlight dates should represent dates, and  or datetime objects.
    highlight_date_labels: list
        The labels for each highlight dates.
    show: bool, optional
        whether or not to immediately show the results

    Returns
    -------
    ax: maptlotlib.Axes object
    """
    daily_trip_counts = G.get_trip_counts_per_day()
    if ax is None:
        _fig, ax = plt.subplots()
    daily_trip_counts["datetime"] = pandas.to_datetime(daily_trip_counts["date_str"])
    daily_trip_counts.plot("datetime", "trip_counts", kind="line", ax=ax, marker="o", color="C0", ls=":",
                           label="Trip counts")
    ax.set_xlabel("Date")
    ax.set_ylabel("Trip counts per day")
    if highlight_dates is not None:
        assert isinstance(highlight_dates, list)
        if highlight_date_labels is not None:
            assert isinstance(highlight_date_labels, list)
            assert len(highlight_dates) == len(highlight_date_labels), "Number of highlight date labels do not match"
        else:
            highlight_date_labels = [None] * len(highlight_dates)
        for i, (highlight_date, label) in enumerate(zip(highlight_dates, highlight_date_labels)):
            color = "C" + str(int(i % 8 + 1))
            highlight_date = pandas.to_datetime(highlight_date)
            ax.axvline(highlight_date, color=color, label=label)
    ax.legend(loc="best")
    ax.grid()
    if show:
        plt.show()
    return ax



from pprint import pformat


class Connection:

    WALK_SEQ = -1
    WALK_TRIP_ID = -1

    def __init__(self, departure_stop, arrival_stop, departure_time, arrival_time, trip_id, seq,
                 is_walk=False, arrival_stop_next_departure_time=float('inf')):
        self.departure_stop = departure_stop
        self.arrival_stop = arrival_stop
        self.departure_time = departure_time
        self.arrival_time = arrival_time
        self.trip_id = trip_id
        self.is_walk = is_walk
        self.seq = int(seq)
        self.arrival_stop_next_departure_time = arrival_stop_next_departure_time

    def duration(self):
        return self.arrival_time - self.departure_time

    def __str__(self):
        return pformat(self.__dict__)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return '<%s:%s:%s:%s:%s:%s:%s:%s>' % (
            self.__class__.__name__, self.departure_stop, self.arrival_stop,
            self.departure_time, self.arrival_time, self.trip_id, self.is_walk, self.arrival_stop_next_departure_time)

    def __hash__(self):
        return hash(self.__repr__())


import os
import sqlite3
import pandas as pd

from gtfspy.routing.connection import Connection
from gtfspy.gtfs import GTFS
from gtfspy.routing.label import LabelTimeAndRoute, LabelTimeWithBoardingsCount, LabelTimeBoardingsAndRoute, \
    compute_pareto_front, LabelGeneric
from gtfspy.routing.travel_impedance_data_store import TravelImpedanceDataStore
from gtfspy.routing.fastest_path_analyzer import FastestPathAnalyzer
from gtfspy.routing.node_profile_analyzer_time_and_veh_legs import NodeProfileAnalyzerTimeAndVehLegs
from gtfspy.util import timeit


def attach_database(conn, other_db_path, name="other"):
    cur = conn.cursor()
    cur.execute("ATTACH '%s' AS '%s'" % (str(other_db_path), name))
    cur.execute("PRAGMA database_list")
    print("other database attached:", cur.fetchall())
    return conn

_T_WALK_STR = "t_walk"

class JourneyDataManager:

    def __init__(self, gtfs_path, journey_db_path, routing_params=None, multitarget_routing=False,
                 track_vehicle_legs=True, track_route=False):
        """
        :param gtfs: GTFS object
        :param list_of_stop_profiles: dict of NodeProfileMultiObjective
        :param multitarget_routing: bool
        """
        self.multitarget_routing = multitarget_routing
        self.track_route = track_route
        self.track_vehicle_legs = track_vehicle_legs
        self.gtfs_path = gtfs_path
        self.gtfs = GTFS(self.gtfs_path)
        self.gtfs_meta = self.gtfs.meta
        self.gtfs._dont_close = True
        self.od_pairs = None
        self._targets = None
        self._origins = None
        self.diff_conn = None

        if not routing_params:
            routing_params = dict()
        self.routing_params_input = routing_params

        assert os.path.exists(journey_db_path) or routing_params is not None
        journey_db_pre_exists = os.path.isfile(journey_db_path)

        # insert a pretty robust timeout:
        timeout = 1000
        self.conn = sqlite3.connect(journey_db_path, timeout)
        if not journey_db_pre_exists:
            self.initialize_database()

        self.routing_parameters = Parameters(self.conn)
        self._assert_journey_computation_paramaters_match()

        self.journey_properties = {"journey_duration": (_T_WALK_STR, _T_WALK_STR)}
        if routing_params.get('track_vehicle_legs', False) or \
                self.routing_parameters.get('track_vehicle_legs', False):
            self.journey_properties["n_boardings"] = (float("inf"), 0)
        if self.track_route:
            additional_journey_parameters = {
                "in_vehicle_duration": (float('inf'), 0),
                "transfer_wait_duration": (float('inf'), 0),
                "walking_duration": (_T_WALK_STR, _T_WALK_STR),
                "pre_journey_wait_fp": (float('inf'), 0)
            }
            self.journey_properties.update(additional_journey_parameters)
        self.travel_impedance_measure_names = list(self.journey_properties.keys())
        self.travel_impedance_measure_names += ["temporal_distance"]

    def __del__(self):
        self.gtfs._dont_close = False
        if self.conn:
            self.conn.close()

    @timeit
    def import_journey_data_for_target_stop(self, target_stop_I, origin_stop_I_to_journey_labels, enforce_synchronous_writes=False):
        """
        Parameters
        ----------
        origin_stop_I_to_journey_labels: dict
            key: origin_stop_Is
            value: list of labels
        target_stop_I: int
        """
        cur = self.conn.cursor()
        self.conn.isolation_level = 'EXCLUSIVE'
        # if not enforce_synchronous_writes:
        cur.execute('PRAGMA synchronous = 0;')

        if self.track_route:
            self._insert_journeys_with_route_into_db(origin_stop_I_to_journey_labels, target_stop=int(target_stop_I))
        else:
            self._insert_journeys_into_db_no_route(origin_stop_I_to_journey_labels, target_stop=int(target_stop_I))
        print("Finished import process")
        self.conn.commit()

    def _assert_journey_computation_paramaters_match(self):
        for key, value in self.routing_parameters.items():
            if key in self.gtfs_meta.keys():
                assert self.gtfs_meta[key] == value

    def _get_largest_journey_id(self):
        cur = self.conn.cursor()
        val = cur.execute("select max(journey_id) FROM journeys").fetchone()
        return val[0] if val[0] else 0

    def _insert_journeys_into_db_no_route(self, stop_profiles, target_stop=None):
        # TODO: Change the insertion so that the check last journey id and insertions are in the same transaction block
        """
        con.isolation_level = 'EXCLUSIVE'
        con.execute('BEGIN EXCLUSIVE')
        #exclusive access starts here. Nothing else can r/w the db, do your magic here.
        con.commit()
        """
        print("Collecting journey data")
        journey_id = 1
        journey_list = []
        tot = len(stop_profiles)
        for i, (origin_stop, labels) in enumerate(stop_profiles.items(), start=1):
            #print("\r Stop " + str(i) + " of " + str(tot), end='', flush=True)
            for label in labels:
                assert (isinstance(label, LabelTimeWithBoardingsCount))
                if self.multitarget_routing:
                    target_stop = None
                else:
                    target_stop = int(target_stop)

                values = [int(journey_id),
                          int(origin_stop),
                          target_stop,
                          int(label.departure_time),
                          int(label.arrival_time_target),
                          int(label.n_boardings)]

                journey_list.append(values)
                journey_id += 1
        print("Inserting journeys without route into database")
        insert_journeys_stmt = '''INSERT INTO journeys(
              journey_id,
              from_stop_I,
              to_stop_I,
              departure_time,
              arrival_time_target,
              n_boardings) VALUES (%s) ''' % (", ".join(["?" for x in range(6)]))
        #self.conn.executemany(insert_journeys_stmt, journey_list)
        self._executemany_exclusive(insert_journeys_stmt, journey_list)
        self.conn.commit()

    @timeit
    def _executemany_exclusive(self, statement, rows):
        self.conn.execute('BEGIN EXCLUSIVE')
        last_id = self._get_largest_journey_id()
        rows = [[x[0] + last_id] + x[1:] for x in rows]
        self.conn.executemany(statement, rows)

    def _insert_journeys_with_route_into_db(self, stop_I_to_journey_labels, target_stop):
        print("Collecting journey and connection data")
        journey_id = (self._get_largest_journey_id() if self._get_largest_journey_id() else 0) + 1
        journey_list = []
        connection_list = []
        label = None
        for i, (origin_stop, labels) in enumerate(stop_I_to_journey_labels.items(), start=1):
            # tot = len(stop_profiles)
            #print("\r Stop " + str(i) + " of " + str(tot), end='', flush=True)

            assert (isinstance(stop_I_to_journey_labels[origin_stop], list))

            for label in labels:
                assert (isinstance(label, LabelTimeAndRoute) or isinstance(label, LabelTimeBoardingsAndRoute))
                # We need to "unpack" the journey to actually figure out where the trip went
                # (there can be several targets).
                if label.departure_time == label.arrival_time_target:
                    print("Weird label:", label)
                    continue

                target_stop, new_connection_values, route_stops = self._collect_connection_data(journey_id, label)
                if origin_stop == target_stop:
                    continue

                if isinstance(label, LabelTimeBoardingsAndRoute):
                    values = [int(journey_id),
                              int(origin_stop),
                              int(target_stop),
                              int(label.departure_time),
                              int(label.arrival_time_target),
                              label.n_boardings,
                              label.movement_duration,
                              route_stops]
                else:
                    values = [int(journey_id),
                              int(origin_stop),
                              int(target_stop),
                              int(label.departure_time),
                              int(label.arrival_time_target),
                              label.movement_duration,
                              route_stops]

                journey_list.append(values)
                connection_list += new_connection_values
                journey_id += 1

        print("Inserting journeys into database")
        if label:
            if isinstance(label, LabelTimeBoardingsAndRoute):
                insert_journeys_stmt = '''INSERT INTO journeys(
                      journey_id,
                      from_stop_I,
                      to_stop_I,
                      departure_time,
                      arrival_time_target,
                      n_boardings,
                      movement_duration,
                      route) VALUES (%s) ''' % (", ".join(["?" for x in range(8)]))
            else:
                insert_journeys_stmt = '''INSERT INTO journeys(
                      journey_id,
                      from_stop_I,
                      to_stop_I,
                      departure_time,
                      arrival_time_target,
                      movement_duration,
                      route) VALUES (%s) ''' % (", ".join(["?" for x in range(7)]))
            self.conn.executemany(insert_journeys_stmt, journey_list)

            print("Inserting legs into database")
            insert_legs_stmt = '''INSERT INTO legs(
                                  journey_id,
                                  from_stop_I,
                                  to_stop_I,
                                  departure_time,
                                  arrival_time_target,
                                  trip_I,
                                  seq,
                                  leg_stops) VALUES (%s) ''' % (", ".join(["?" for x in range(8)]))
            self.conn.executemany(insert_legs_stmt, connection_list)
            self.routing_parameters["target_list"] += (str(target_stop) + ",")
            self.conn.commit()


    def create_index_for_journeys_table(self):
        self.conn.execute("PRAGMA temp_store=2")
        self.conn.commit()
        self.conn.execute("CREATE INDEX IF NOT EXISTS journeys_to_stop_I_idx ON journeys (to_stop_I)")

    def _collect_connection_data(self, journey_id, label):
        target_stop = None
        cur_label = label
        seq = 1
        value_list = []
        route_stops = []
        leg_stops = []
        prev_trip_id = None
        connection = None
        leg_departure_time = None
        leg_departure_stop = None
        leg_arrival_time = None
        leg_arrival_stop = None
        while True:
            if isinstance(cur_label.connection, Connection):
                connection = cur_label.connection
                if connection.trip_id:
                    trip_id = connection.trip_id
                else:
                    trip_id = -1

                # In case of new leg
                if prev_trip_id != trip_id:
                    route_stops.append(connection.departure_stop)
                    if prev_trip_id:
                        leg_stops.append(connection.departure_stop)

                        values = (
                            int(journey_id),
                            int(leg_departure_stop),
                            int(leg_arrival_stop),
                            int(leg_departure_time),
                            int(leg_arrival_time),
                            int(prev_trip_id),
                            int(seq),
                            ','.join([str(x) for x in leg_stops])
                                )
                        value_list.append(values)
                        seq += 1
                        leg_stops = []

                    leg_departure_stop = connection.departure_stop
                    leg_departure_time = connection.departure_time
                leg_arrival_time = connection.arrival_time
                leg_arrival_stop = connection.arrival_stop
                leg_stops.append(connection.departure_stop)
                target_stop = connection.arrival_stop
                prev_trip_id = trip_id

            if not cur_label.previous_label:
                leg_stops.append(connection.arrival_stop)
                values = (
                    int(journey_id),
                    int(leg_departure_stop),
                    int(leg_arrival_stop),
                    int(leg_departure_time),
                    int(leg_arrival_time),
                    int(prev_trip_id),
                    int(seq),
                    ','.join([str(x) for x in leg_stops])
                )
                value_list.append(values)
                break

            cur_label = cur_label.previous_label
        route_stops.append(target_stop)
        route_stops = ','.join([str(x) for x in route_stops])
        return target_stop, value_list, route_stops

    def populate_additional_journey_columns(self):
        self.add_fastest_path_column()
        self.add_time_to_prev_journey_fp_column()
        self.compute_journey_time_components()
        self.calculate_pre_journey_waiting_times_ignoring_direct_walk()

    def get_od_pairs_having_journeys(self):
        cur = self.conn.cursor()
        if not self.od_pairs:
            cur.execute('SELECT from_stop_I, to_stop_I FROM journeys GROUP BY from_stop_I, to_stop_I')
            self.od_pairs = cur.fetchall()
        return self.od_pairs

    def get_targets_having_journeys(self):
        cur = self.conn.cursor()
        if not self._targets:
            cur.execute('SELECT to_stop_I FROM journeys GROUP BY to_stop_I')
            self._targets = [target[0] for target in cur.fetchall()]
        return self._targets

    def get_origins_having_journeys(self):
        cur = self.conn.cursor()
        if not self._origins:
            cur.execute('SELECT from_stop_I FROM journeys GROUP BY from_stop_I')
            self._origins = [origin[0] for origin in cur.fetchall()]
        return self._origins

    def get_table_with_coordinates(self, table_name, target=None):
        df = self.get_table_as_dataframe(table_name, target)
        return self.gtfs.add_coordinates_to_df(df, join_column='from_stop_I')

    def get_table_as_dataframe(self, table_name, to_stop_I_target=None):
        query = "SELECT * FROM " + table_name
        if to_stop_I_target:
            query += " WHERE to_stop_I = %s" % to_stop_I_target
        return pd.read_sql_query(query, self.conn)

    @timeit
    def add_fastest_path_column(self):
        print("adding fastest path column")
        cur = self.conn.cursor()
        for target in self.get_targets_having_journeys():
            fastest_path_journey_ids = []
            for origin in self.get_origins_having_journeys():
                cur.execute('SELECT departure_time, arrival_time_target, journey_id FROM journeys '
                            'WHERE from_stop_I = ? AND to_stop_I = ? '
                            'ORDER BY departure_time ASC', (origin, target))
                all_trips = cur.fetchall()
                all_labels = [LabelTimeAndRoute(x[0], x[1], x[2], False) for x in all_trips] #putting journey_id as movement_duration
                all_fp_labels = compute_pareto_front(all_labels, finalization=False, ignore_n_boardings=True)
                fastest_path_journey_ids.append(all_fp_labels)

            fastest_path_journey_ids = [(1, x.movement_duration) for sublist in fastest_path_journey_ids for x in sublist]
            cur.executemany("UPDATE journeys SET fastest_path = ? WHERE journey_id = ?", fastest_path_journey_ids)
        self.conn.commit()

    @timeit
    def add_time_to_prev_journey_fp_column(self):
        print("adding pre journey waiting time")
        cur = self.conn.cursor()
        for target in self.get_targets_having_journeys():

            cur.execute('SELECT journey_id, from_stop_I, to_stop_I, departure_time FROM journeys '
                        'WHERE fastest_path = 1 AND to_stop_I = ? '
                        'ORDER BY from_stop_I, to_stop_I, departure_time ', (target,))

            all_trips = cur.fetchall()
            time_to_prev_journey = []
            prev_departure_time = None
            prev_origin = None
            prev_destination = None
            for trip in all_trips:
                journey_id = trip[0]
                from_stop_I = trip[1]
                to_stop_I = trip[2]
                departure_time = trip[3]
                if prev_origin != from_stop_I or prev_destination != to_stop_I:
                    prev_departure_time = None
                if prev_departure_time:
                    time_to_prev_journey.append((departure_time - prev_departure_time, journey_id))
                prev_origin = from_stop_I
                prev_destination = to_stop_I
                prev_departure_time = departure_time
            cur.executemany("UPDATE journeys SET pre_journey_wait_fp = ? WHERE journey_id = ?", time_to_prev_journey)
        self.conn.commit()

    @timeit
    def compute_journey_time_components(self):
        print("adding journey components")
        cur = self.conn.cursor()
        cur.execute("UPDATE journeys SET journey_duration = arrival_time_target - departure_time")

        if self.track_route:
            cur.execute("UPDATE journeys "
                        "SET "
                        "in_vehicle_duration = "
                        "(SELECT sum(arrival_time_target - departure_time) AS in_vehicle_duration FROM legs "
                        "WHERE journeys.journey_id = legs.journey_id AND trip_I != -1 GROUP BY journey_id)")
            cur.execute("UPDATE journeys "
                        "SET "
                        "walking_duration = "
                        "(SELECT sum(arrival_time_target - departure_time) AS walking_duration FROM legs "
                        "WHERE journeys.journey_id = legs.journey_id AND trip_I < 0 GROUP BY journey_id)")
            cur.execute("UPDATE journeys "
                        "SET transfer_wait_duration = journey_duration - in_vehicle_duration - walking_duration")
        self.conn.commit()

    def _journey_label_generator(self, destination_stop_Is=None, origin_stop_Is=None):
        """
        Parameters
        ----------
        destination_stop_Is: list-like
        origin_stop_Is: list-like

        Yields
        ------
        (origin_stop_I, destination_stop_I, journey_labels) : tuple
        """
        conn = self.conn
        conn.row_factory = sqlite3.Row
        if destination_stop_Is is None:
            destination_stop_Is = self.get_targets_having_journeys()
        if origin_stop_Is is None:
            origin_stop_Is = self.get_origins_having_journeys()

        for destination_stop_I in destination_stop_Is:
            if self.track_route:
                label_features = "journey_id, from_stop_I, to_stop_I, n_boardings, movement_duration, " \
                                 "journey_duration, in_vehicle_duration, transfer_wait_duration, walking_duration, " \
                                 "departure_time, arrival_time_target"""
            else:
                label_features = "journey_id, from_stop_I, to_stop_I, n_boardings, departure_time, " \
                                 "arrival_time_target"
            sql = "SELECT " + label_features + " FROM journeys WHERE to_stop_I = %s" % destination_stop_I

            df = pd.read_sql_query(sql, self.conn)
            for origin_stop_I in origin_stop_Is:
                selection = df.loc[df['from_stop_I'] == origin_stop_I]
                journey_labels = []
                for journey in selection.to_dict(orient='records'):
                    journey["pre_journey_wait_fp"] = -1
                    try:
                        journey_labels.append(LabelGeneric(journey))
                    except Exception as e:
                        print(journey)
                        raise e
                yield origin_stop_I, destination_stop_I, journey_labels

    def get_node_profile_time_analyzer(self, target, origin, start_time_dep, end_time_dep):
        sql = """SELECT journey_id, from_stop_I, to_stop_I, n_boardings, movement_duration, journey_duration,
        in_vehicle_duration, transfer_wait_duration, walking_duration, departure_time, arrival_time_target
        FROM journeys WHERE to_stop_I = %s AND from_stop_I = %s""" % (target, origin)
        df = pd.read_sql_query(sql, self.conn)
        journey_labels = []
        for journey in df.to_dict(orient='records'):
            journey_labels.append(LabelGeneric(journey))

        fpa = FastestPathAnalyzer(journey_labels,
                                  start_time_dep,
                                  end_time_dep,
                                  walk_duration=float('inf'),  # walking time
                                  label_props_to_consider=list(self.journey_properties.keys()))
        return fpa.get_time_analyzer()

    def get_node_profile_analyzer_time_and_veh_legs(self, target, origin, start_time_dep, end_time_dep):
        sql = """SELECT from_stop_I, to_stop_I, n_boardings, departure_time, arrival_time_target FROM journeys WHERE to_stop_I = %s AND from_stop_I = %s""" % (target, origin)
        df = pd.read_sql_query(sql, self.conn)

        journey_labels = []
        for journey in df.itertuples():
            departure_time = journey.departure_time
            arrival_time_target = journey.arrival_time_target
            n_boardings = journey.n_boardings
            journey_labels.append(LabelTimeWithBoardingsCount(departure_time,
                                                              arrival_time_target,
                                                              n_boardings,
                                                              first_leg_is_walk=float('nan')))

        # This ought to be optimized...
        query = """SELECT d, d_walk FROM stop_distances WHERE to_stop_I = %s AND from_stop_I = %s""" % (target, origin)
        df = self.gtfs.execute_custom_query_pandas(query)
        if len(df) > 0:
            walk_duration = float(df['d_walk']) / self.routing_params_input['walk_speed']
        else:
            walk_duration = float('inf')
        analyzer = NodeProfileAnalyzerTimeAndVehLegs(journey_labels,
                                                     walk_duration,  # walking time
                                                     start_time_dep,
                                                     end_time_dep)
        return analyzer

    def __compute_travel_impedance_measure_dict(self,
                                                origin,
                                                target,
                                                journey_labels,
                                                analysis_start_time,
                                                analysis_end_time):
        measure_summaries = {}
        kwargs = {"from_stop_I": origin, "to_stop_I": target}
        walking_distance = self.gtfs.get_stop_distance(origin, target)

        if walking_distance:
            walking_duration = walking_distance / self.routing_params_input["walk_speed"]
        else:
            walking_duration = float("inf")
        fpa = FastestPathAnalyzer(journey_labels,
                                  analysis_start_time,
                                  analysis_end_time,
                                  walk_duration=walking_duration,  # walking time
                                  label_props_to_consider=list(self.journey_properties.keys()),
                                  **kwargs)
        temporal_distance_analyzer = fpa.get_temporal_distance_analyzer()
        # Note: the summary_as_dict automatically includes also the from_stop_I and to_stop_I -fields.
        measure_summaries["temporal_distance"] = temporal_distance_analyzer.summary_as_dict()
        fpa.calculate_pre_journey_waiting_times_ignoring_direct_walk()
        for key, (value_no_next_journey, value_cutoff) in self.journey_properties.items():
            value_cutoff = walking_duration if value_cutoff == _T_WALK_STR else value_cutoff
            value_no_next_journey = walking_duration if value_no_next_journey == _T_WALK_STR else value_no_next_journey
            if key == "pre_journey_wait_fp":
                property_analyzer = fpa.get_prop_analyzer_for_pre_journey_wait()
            else:
                property_analyzer = fpa.get_prop_analyzer_flat(key, value_no_next_journey, value_cutoff)
            measure_summaries[key] = property_analyzer.summary_as_dict()
        return measure_summaries

    def compute_travel_impedance_measures_for_target(self,
                                                     analysis_start_time,
                                                     analysis_end_time,
                                                     target, origins=None):
        if origins is None:
            origins = self.get_origins_having_journeys()
        measure_to_measure_summary_dicts = {}
        for measure in ["temporal_distance"] + list(self.journey_properties):
            measure_to_measure_summary_dicts[measure] = []
        for origin, target, journey_labels in self._journey_label_generator([target], origins):
            measure_summary_dicts_for_pair = \
            self.__compute_travel_impedance_measure_dict(
                origin, target, journey_labels,
                analysis_start_time, analysis_end_time
            )
            for measure in measure_summary_dicts_for_pair:
                measure_to_measure_summary_dicts[measure].append(measure_summary_dicts_for_pair[measure])
        return measure_to_measure_summary_dicts

    @timeit
    def compute_and_store_travel_impedance_measures(self,
                                                    analysis_start_time,
                                                    analysis_end_time,
                                                    travel_impedance_store_fname,
                                                    origins=None,
                                                    targets=None):

        data_store = TravelImpedanceDataStore(travel_impedance_store_fname)

        measure_to_measure_summary_dicts = {}
        for travel_impedance_measure in self.travel_impedance_measure_names:
            data_store.create_table(travel_impedance_measure)

        print("Computing total number of origins and targets..", end='', flush=True)
        if targets is None:
            targets = self.get_targets_having_journeys()
        if origins is None:
            origins = self.get_origins_having_journeys()
        print("\rComputed total number of origins and targets")
        n_pairs_tot = len(origins) * len(targets)

        print("TODO!, compute_and_store_travel_impedance_measures, "
              "should be adjusted to use "
              "travel_impedance_measure_data_store.py instead")

        def _flush_data_to_db(results):
            for travel_impedance_measure, data in results.items():
                data_store.insert_data(travel_impedance_measure, data)
            for travel_impedance_measure in self.travel_impedance_measure_names:
                results[travel_impedance_measure] = []

        # This initializes the meaasure_to_measure_summary_dict properly
        _flush_data_to_db(measure_to_measure_summary_dicts)

        for i, (origin, target, journey_labels) in enumerate(self._journey_label_generator(targets, origins)):
            print("i", origin, target, journey_labels)
            if len(journey_labels) == 0:
                continue

            measure_summary_dicts_for_pair = self.__compute_travel_impedance_measure_dict(origin, target, journey_labels,
                                                                                          analysis_start_time, analysis_end_time)
            for measure in measure_summary_dicts_for_pair:
                measure_to_measure_summary_dicts[measure].append(measure_summary_dicts_for_pair[measure])

            if i % 1000 == 0: # update in batches of 1000
                print("\r", i, "/", n_pairs_tot, " : ", "%.2f" % round(float(i) / n_pairs_tot, 3), end='', flush=True)
                _flush_data_to_db(measure_to_measure_summary_dicts)

        # flush everything that remains
        _flush_data_to_db(measure_to_measure_summary_dicts)

    @timeit
    def calculate_pre_journey_waiting_times_ignoring_direct_walk(self):
        all_fp_labels = []
        for origin, target, journey_labels in self._journey_label_generator():
            if not journey_labels:
                continue
            fpa = FastestPathAnalyzer(journey_labels,
                                      self.routing_parameters["routing_start_time_dep"],
                                      self.routing_parameters["routing_end_time_dep"],
                                      walk_duration=float('inf'))
            fpa.calculate_pre_journey_waiting_times_ignoring_direct_walk()
            all_fp_labels += fpa.get_fastest_path_labels()
        self.update_journey_from_labels(all_fp_labels, "pre_journey_wait_fp")

    def update_journey_from_labels(self, labels, attribute):
        cur = self.conn.cursor()
        insert_tuples = []
        for label in labels:
            insert_tuples.append((getattr(label, attribute), getattr(label, "journey_id")))

        sql = "UPDATE journeys SET %s = ? WHERE journey_id = ?" % (attribute,)
        cur.executemany(sql, insert_tuples)
        self.conn.commit()

    def _insert_travel_impedance_data_to_db(self, travel_impedance_measure_name, data):
        """
        Parameters
        ----------
        travel_impedance_measure_name: str
        data: list[dict]
            Each list element must contain keys:
            "from_stop_I", "to_stop_I", "min", "max", "median" and "mean"
        """
        f = float
        data_tuple = [(x["from_stop_I"], x["to_stop_I"], f(x["min"]), f(x["max"]), f(x["median"]), f(x["mean"])) for x in data]
        insert_stmt = '''INSERT OR REPLACE INTO ''' + travel_impedance_measure_name + ''' (
                              from_stop_I,
                              to_stop_I,
                              min,
                              max,
                              median,
                              mean) VALUES (?, ?, ?, ?, ?, ?) '''
        self.conn.executemany(insert_stmt, data_tuple)
        self.conn.commit()

    def _create_index_for_journeys_table(self):
        self.conn.execute("CREATE INDEX IF NOT EXISTS journeys_to_stop_I_idx ON journeys (to_stop_I)")

    @timeit
    def initialize_comparison_tables(self, diff_db_path, before_db_tuple, after_db_tuple):
        self.diff_conn = sqlite3.connect(diff_db_path)

        self.diff_conn = attach_database(self.diff_conn, before_db_tuple[0], name=before_db_tuple[1])
        self.diff_conn = attach_database(self.diff_conn, after_db_tuple[0], name=after_db_tuple[1])

        for table in self.travel_impedance_measure_names:
            self.diff_conn.execute("CREATE TABLE IF NOT EXISTS diff_" + table +
                                   " (from_stop_I, to_stop_I, diff_min, diff_max, diff_median, diff_mean)")
            insert_stmt = "INSERT OR REPLACE INTO diff_" + table + \
                          "(from_stop_I, to_stop_I, diff_min, diff_max, diff_median, diff_mean) " \
                          "SELECT t1.from_stop_I, t1.to_stop_I, " \
                          "t1.min - t2.min AS diff_min, " \
                          "t1.max - t2.max AS diff_max, " \
                          "t1.median - t2.median AS diff_median, " \
                          "t1.mean - t2.mean AS diff_mean " \
                          "FROM " + before_db_tuple[1] + "." + table + " AS t1, " \
                          + before_db_tuple[1] + "." + table + " AS t2 " \
                                                               "WHERE t1.from_stop_I = t2.from_stop_I " \
                                                               "AND t1.to_stop_I = t2.to_stop_I "
            self.diff_conn.execute(insert_stmt)
            self.diff_conn.commit()

    def initialize_database(self):
        self._set_up_database()
        self._initialize_parameter_table()
        print("Database initialized!")

    def _set_up_database(self):
        self.conn.execute('''CREATE TABLE IF NOT EXISTS parameters(
                             key TEXT UNIQUE,
                             value BLOB)''')
        if self.track_route:
            self.conn.execute('''CREATE TABLE IF NOT EXISTS journeys(
                             journey_id INTEGER PRIMARY KEY,
                             from_stop_I INT,
                             to_stop_I INT,
                             departure_time INT,
                             arrival_time_target INT,
                             n_boardings INT,
                             movement_duration INT,
                             route TEXT,
                             journey_duration INT,
                             pre_journey_wait_fp INT,
                             in_vehicle_duration INT,
                             transfer_wait_duration INT,
                             walking_duration INT,
                             fastest_path INT)''')

            self.conn.execute('''CREATE TABLE IF NOT EXISTS legs(
                         journey_id INT,
                         from_stop_I INT,
                         to_stop_I INT,
                         departure_time INT,
                         arrival_time_target INT,
                         trip_I INT,
                         seq INT,
                         leg_stops TEXT)''')
            """
            self.conn.execute('''CREATE TABLE IF NOT EXISTS nodes(
                         stop_I INT,
                         agg_temp_distances REAL,
                         agg_journey_duration REAL,
                         agg_boardings REAL,
                         agg_transfer_wait REAL,
                         agg_pre_journey_wait REAL,
                         agg_walking_duration REAL)''')

            self.conn.execute('''CREATE TABLE IF NOT EXISTS od_pairs(
                         from_stop_I INT,
                         to_stop_I INT,
                         avg_temp_distance REAL,
                         agg_journey_duration REAL,
                         agg_boardings REAL,
                         agg_transfer_wait REAL,
                         agg_pre_journey_wait REAL,
                         agg_walking_duration REAL)''')


            self.conn.execute('''CREATE TABLE IF NOT EXISTS sections(
                         from_stop_I INT,
                         to_stop_I INT,
                         from_stop_pair_I INT,
                         to_stop_pair_I INT,
                         avg_temp_distance INT,
                         avg_journey_duration INT,
                         n_trips INT)''')

            self.conn.execute('''CREATE TABLE IF NOT EXISTS transfer_nodes(
                         from_stop_I INT,
                         to_stop_I INT,
                         from_stop_pair_I INT,
                         to_stop_pair_I INT,
                         avg_waiting_time INT,
                         n_trips INT)''')
            """
        else:
            self.conn.execute('''CREATE TABLE IF NOT EXISTS journeys(
                         journey_id INTEGER PRIMARY KEY,
                         from_stop_I INT,
                         to_stop_I INT,
                         departure_time INT,
                         arrival_time_target INT,
                         n_boardings INT,
                         journey_duration INT,
                         time_to_prev_journey_fp INT,
                         fastest_path INT)''')

        self.conn.commit()

    def _initialize_parameter_table(self):

        parameters = Parameters(self.conn)

        parameters["multiple_targets"] = self.multitarget_routing
        parameters["gtfs_dir"] = self.gtfs_path
        for param in ["location_name",
                      "lat_median",
                      "lon_median",
                      "start_time_ut",
                      "end_time_ut",
                      "start_date",
                      "end_date"]:
            parameters[param] = self.gtfs_meta[param]
        parameters["target_list"] = ","
        for key, value in self.routing_params_input.items():
            parameters[key] = value
        self.conn.commit()

    def create_indices(self):
        self.conn.execute("PRAGMA temp_store=2")
        self.conn.commit()
        # Next 3 lines are python 3.6 work-arounds again.
        self.conn.isolation_level = None  # former default of autocommit mode
        cur = self.conn.cursor()
        cur.execute('VACUUM;')
        self.conn.isolation_level = ''  # back to python default
        # end python3.6 workaround
        print("Analyzing...")
        cur.execute('ANALYZE')
        print("Indexing")
        cur = self.conn.cursor()
        cur.execute('CREATE INDEX IF NOT EXISTS idx_journeys_jid ON journeys (journey_id)')
        cur.execute('CREATE INDEX IF NOT EXISTS idx_journeys_fid ON journeys (from_stop_I)')
        cur.execute('CREATE INDEX IF NOT EXISTS idx_journeys_tid ON journeys (to_stop_I)')

        if self.track_route:
            cur.execute('CREATE INDEX IF NOT EXISTS idx_legs_jid ON legs (journey_id)')
            cur.execute('CREATE INDEX IF NOT EXISTS idx_legs_trid ON legs (trip_I)')
            cur.execute('CREATE INDEX IF NOT EXISTS idx_legs_fid ON legs (from_stop_I)')
            cur.execute('CREATE INDEX IF NOT EXISTS idx_legs_tid ON legs (to_stop_I)')
            cur.execute('CREATE INDEX IF NOT EXISTS idx_journeys_route ON journeys (route)')
        self.conn.commit()


class DiffDataManager:
    def __init__(self, diff_db_path):
        self.conn = sqlite3.connect(diff_db_path)

    def initialize_journey_comparison_tables(self, tables, before_db_tuple, after_db_tuple):
        before_db_path = before_db_tuple[0]
        before_db_name = before_db_tuple[1]
        after_db_path = after_db_tuple[0]
        after_db_name = after_db_tuple[1]

        self.conn = self.attach_database(before_db_path, name=before_db_name)
        self.conn = self.attach_database(after_db_path, name=after_db_name)

        for table in tables:
            self.conn.execute("CREATE TABLE IF NOT EXISTS diff_" + table +
                              "(from_stop_I INT, to_stop_I INT, "
                              "diff_min INT, diff_max INT, diff_median INT, diff_mean INT, "
                              "rel_diff_min REAL, rel_diff_max REAL, rel_diff_median REAL, rel_diff_mean REAL)")
            insert_stmt = "INSERT OR REPLACE INTO diff_" + table + \
                          " (from_stop_I, to_stop_I, diff_min, diff_max, diff_median, diff_mean, " \
                          "rel_diff_min, rel_diff_max, rel_diff_median, rel_diff_mean) " \
                          "SELECT " \
                          "t1.from_stop_I, " \
                          "t1.to_stop_I, " \
                          "t1.min - t2.min AS diff_min, " \
                          "t1.max - t2.max AS diff_max, " \
                          "t1.median - t2.median AS diff_median, " \
                          "t1.mean - t2.mean AS diff_mean, " \
                          "(t1.min - t2.min)*1.0/t2.min AS rel_diff_min, " \
                          "(t1.max - t2.max)*1.0/t2.max AS rel_diff_max, " \
                          "(t1.median - t2.median)*1.0/t2.median AS rel_diff_median, " \
                          "(t1.mean - t2.mean)*1.0/t2.mean AS rel_diff_mean " \
                          "FROM " + after_db_name + "." + table + " AS t1, "\
                          + before_db_name + "." + table + \
                          " AS t2 WHERE t1.from_stop_I = t2.from_stop_I AND t1.to_stop_I = t2.to_stop_I "
            self.conn.execute(insert_stmt)
            self.conn.commit()

    def attach_database(self, other_db_path, name="other"):
        cur = self.conn.cursor()
        cur.execute("ATTACH '%s' AS '%s'" % (str(other_db_path), name))
        cur.execute("PRAGMA database_list")
        print("other database attached:", cur.fetchall())
        return self.conn

    def get_table_with_coordinates(self, gtfs, table_name, target=None, use_relative=False):
        df = self.get_table_as_dataframe(table_name, use_relative, target)
        return gtfs.add_coordinates_to_df(df, join_column='from_stop_I')

    def get_table_as_dataframe(self, table_name, use_relative, target=None):
        if use_relative:
            query = "SELECT from_stop_I, to_stop_I, rel_diff_min, rel_diff_max, rel_diff_median, rel_diff_mean FROM "\
                    + table_name
        else:
            query = "SELECT from_stop_I, to_stop_I, diff_min, diff_max, diff_median, diff_mean FROM " + table_name
        if target:
            query += " WHERE to_stop_I = %s" % target
        return pd.read_sql_query(query, self.conn)

    def get_temporal_distance_change_o_d_pairs(self, target, threshold):
        cur = self.conn.cursor()
        query = """SELECT from_stop_I FROM diff_temporal_distance
                    WHERE to_stop_I = %s AND abs(diff_mean) >= %s""" % (target, threshold)
        rows = [x[0] for x in cur.execute(query).fetchall()]
        return rows

    def get_largest_component(self, target, threshold=180):
        query = """SELECT diff_pre_journey_wait_fp.from_stop_I AS stop_I, 
                    diff_pre_journey_wait_fp.diff_mean AS pre_journey_wait, 
                    diff_in_vehicle_duration.diff_mean AS in_vehicle_duration,
                    diff_transfer_wait_duration.diff_mean AS transfer_wait, 
                    diff_walking_duration.diff_mean AS walking_duration,
                    diff_temporal_distance.diff_mean AS temporal_distance
                    FROM diff_pre_journey_wait_fp, diff_in_vehicle_duration, 
                    diff_transfer_wait_duration, diff_walking_duration, diff_temporal_distance
                    WHERE diff_pre_journey_wait_fp.rowid = diff_in_vehicle_duration.rowid
                    AND diff_pre_journey_wait_fp.rowid = diff_transfer_wait_duration.rowid
                    AND diff_pre_journey_wait_fp.rowid = diff_walking_duration.rowid
                    AND diff_pre_journey_wait_fp.rowid = diff_temporal_distance.rowid
                    AND diff_pre_journey_wait_fp.to_stop_I = %s""" % (target,)
        df = pd.read_sql_query(query, self.conn)
        df['max_component'] = df[["pre_journey_wait", "in_vehicle_duration", "transfer_wait", "walking_duration"]].idxmax(axis=1)
        df['max_value'] = df[["pre_journey_wait", "in_vehicle_duration", "transfer_wait", "walking_duration"]].max(axis=1)

        mask = (df['max_value'] < threshold)

        df.loc[mask, 'max_component'] = "no_change_within_threshold"

        df['min_component'] = df[["pre_journey_wait", "in_vehicle_duration", "transfer_wait", "walking_duration"]].idxmin(axis=1)
        df['min_value'] = df[["pre_journey_wait", "in_vehicle_duration", "transfer_wait", "walking_duration"]].min(axis=1)

        mask = (df['min_value'] > -1 * threshold)

        df.loc[mask, 'min_component'] = "no_change_within_threshold"

        return df


class Parameters(object):
    """
    This provides dictionary protocol for updating parameters table, similar to GTFS metadata ("meta table").
    """

    def __init__(self, conn):
        self._conn = conn
        self._conn.execute("CREATE TABLE IF NOT EXISTS parameters (key, value)")

    def __setitem__(self, key, value):
        self._conn.execute("INSERT OR REPLACE INTO parameters('key', 'value') VALUES (?, ?)", (key, value))
        self._conn.commit()

    def __getitem__(self, key):
        cur = self._conn.cursor()
        cur.execute("SELECT value FROM parameters WHERE key=?", (key,))
        val = cur.fetchone()
        if not val:
            raise KeyError("This journey db does not have parameter: %s" % key)
        return val[0]

    def __delitem__(self, key):
        self._conn.execute("DELETE FROM parameters WHERE key=?", (key,))
        self._conn.commit()

    def __iter__(self):
        cur = self._conn.execute('SELECT key FROM parameters ORDER BY key')
        return (x[0] for x in cur)

    def __contains__(self, key):
        val = self._conn.execute('SELECT value FROM parameters WHERE key=?',
                                 (key,)).fetchone()
        return val is not None

    def get(self, key, default=None):
        val = self._conn.execute('SELECT value FROM parameters WHERE key=?',
                                 (key,)).fetchone()
        if not val:
            return default
        return val[0]

    def items(self):
        cur = self._conn.execute('SELECT key, value FROM parameters ORDER BY key')
        return cur

    def keys(self):
        cur = self._conn.execute('SELECT key FROM parameters ORDER BY key')
        return cur

    def values(self):
        cur = self._conn.execute('SELECT value FROM parameters ORDER BY key')
        return cur



import copy

from gtfspy.routing.label import LabelTime, compute_pareto_front_naive


class NodeProfileC:
    """
    In the connection scan algorithm, each stop has a profile entry
    that stores information on the Pareto-Optimal (departure_time_this_node, arrival_time_target_node) tuples.
    """

    def __init__(self, walk_to_target_duration=float('inf')):
        self._labels = []  # list[LabelTime] # always ordered by decreasing departure_time
        self._walk_to_target_duration = walk_to_target_duration

    def get_walk_to_target_duration(self):
        return self._walk_to_target_duration

    def update_pareto_optimal_tuples(self, new_label):
        """
        Parameters
        ----------
        new_label: LabelTime

        Returns
        -------
        updated: bool
        """
        assert (isinstance(new_label, LabelTime))
        if self._labels:
            assert (new_label.departure_time <= self._labels[-1].departure_time)
            best_later_departing_arrival_time = self._labels[-1].arrival_time_target
        else:
            best_later_departing_arrival_time = float('inf')

        walk_to_target_arrival_time = new_label.departure_time + self._walk_to_target_duration

        best_arrival_time = min(walk_to_target_arrival_time,
                                best_later_departing_arrival_time,
                                new_label.arrival_time_target)
        # this should be changed to get constant time insertions / additions
        # (with time-indexing)
        if (new_label.arrival_time_target < walk_to_target_arrival_time and
                new_label.arrival_time_target < best_later_departing_arrival_time):
            self._labels.append(LabelTime(new_label.departure_time, best_arrival_time))
            return True
        else:
            return False

    def evaluate_earliest_arrival_time_at_target(self, dep_time, transfer_margin):
        """
        Get the earliest arrival time at the target, given a departure time.

        Parameters
        ----------
        dep_time : float, int
            time in unix seconds
        transfer_margin: float, int
            transfer margin in seconds

        Returns
        -------
        arrival_time : float
            Arrival time in the given time unit (seconds after unix epoch).
        """
        minimum = dep_time + self._walk_to_target_duration
        for label in self._labels[::-1]:
            if label.departure_time >= dep_time + transfer_margin:
                minimum = min(minimum, label.arrival_time_target)
                break
        return float(minimum)

    def get_final_optimal_labels(self):
        non_walk_valid_labels = []
        for label in self._labels:
            if label.duration() < self._walk_to_target_duration:
                non_walk_valid_labels.append(label)
        return [label.get_copy() for label in compute_pareto_front_naive(non_walk_valid_labels)]



import numpy

from gtfspy.routing.label import LabelTimeWithBoardingsCount, merge_pareto_frontiers, compute_pareto_front, \
    LabelVehLegCount, LabelTime, LabelTimeBoardingsAndRoute, LabelTimeAndRoute
from gtfspy.routing.connection import Connection


class NodeProfileMultiObjective:
    """
    In the multi-objective connection scan algorithm,
    each stop has a profile entry containing all Pareto-optimal entries.
    """

    def __init__(self,
                 dep_times=None,
                 walk_to_target_duration=float('inf'),
                 label_class=LabelTimeWithBoardingsCount,
                 transit_connection_dep_times=None,
                 closest_target=None,
                 node_id=None):
        """
        Parameters
        ----------
        dep_times
        walk_to_target_duration
        label_class: label class to be used
        transit_connection_dep_times:
            if not given, all connections are assumed to be real connections
        closest_target: int, optional
            stop_I of the closest target if within walking distance (and Routes are recorded)
        """

        if dep_times is None:
            dep_times = []
        n_dep_times = len(dep_times)
        assert n_dep_times == len(set(dep_times)), "There should be no duplicate departure times"
        self._departure_times = list(reversed(sorted(dep_times)))
        self.dep_times_to_index = dict(zip(self._departure_times, range(len(self._departure_times))))
        self._label_bags = [[]] * len(self._departure_times)
        self._walk_to_target_duration = walk_to_target_duration
        self._min_dep_time = float('inf')
        self.label_class = label_class
        self.closest_target = closest_target
        if self.label_class == LabelTimeBoardingsAndRoute and self._walk_to_target_duration < float('inf'):
            assert (self.closest_target is not None)

        if transit_connection_dep_times is not None:
            self._connection_dep_times = transit_connection_dep_times
        else:
            self._connection_dep_times = dep_times
        assert (isinstance(self._connection_dep_times, (list, numpy.ndarray)))
        self._closed = False
        self._finalized = False
        self._final_pareto_optimal_labels = None
        self._real_connection_labels = None
        self.node_id = node_id

    def _check_dep_time_is_valid(self, dep_time):
        """
        A simple checker, that connections are coming in descending order of departure time
        and that no departure time has been "skipped".

        Parameters
        ----------
        dep_time

        Returns
        -------
        None
        """
        assert dep_time <= self._min_dep_time, "Labels should be entered in decreasing order of departure time."
        dep_time_index = self.dep_times_to_index[dep_time]
        if self._min_dep_time < float('inf'):
            min_dep_index = self.dep_times_to_index[self._min_dep_time]
            assert min_dep_index == dep_time_index or (min_dep_index == dep_time_index - 1), \
                "dep times should be ordered sequentially"
        else:
            assert dep_time_index is 0, "first dep_time index should be zero (ensuring that all connections are properly handled)"
        self._min_dep_time = dep_time

    def get_walk_to_target_duration(self):
        """
        Get walking distance to target node.

        Returns
        -------
        walk_to_target_duration: float
        """
        return self._walk_to_target_duration

    def update(self, new_labels, departure_time_backup=None):
        """
        Update the profile with the new labels.
        Each new label should have the same departure_time.

        Parameters
        ----------
        new_labels: list[LabelTime]

        Returns
        -------
        added: bool
            whether new_pareto_tuple was added to the set of pareto-optimal tuples
        """
        if self._closed:
            raise RuntimeError("Profile is closed, no updates can be made")
        try:
            departure_time = next(iter(new_labels)).departure_time
        except StopIteration:
            departure_time = departure_time_backup
        self._check_dep_time_is_valid(departure_time)

        for new_label in new_labels:
            assert (new_label.departure_time == departure_time)
        dep_time_index = self.dep_times_to_index[departure_time]

        if dep_time_index > 0:
            # Departure time is modified in order to not pass on labels which are not Pareto-optimal when departure time is ignored.
            mod_prev_labels = [label.get_copy_with_specified_departure_time(departure_time) for label
                               in self._label_bags[dep_time_index - 1]]
        else:
            mod_prev_labels = list()
        mod_prev_labels += self._label_bags[dep_time_index]

        walk_label = self._get_label_to_target(departure_time)
        if walk_label:
            new_labels = new_labels + [walk_label]
        new_frontier = merge_pareto_frontiers(new_labels, mod_prev_labels)

        self._label_bags[dep_time_index] = new_frontier
        return True

    def evaluate(self, dep_time, first_leg_can_be_walk=True, connection_arrival_time=None):

        """
        Get the pareto_optimal set of Labels, given a departure time.

        Parameters
        ----------
        dep_time : float, int
            time in unix seconds
        first_leg_can_be_walk : bool, optional
            whether to allow walking to target to be included into the profile
            (I.e. whether this function is called when scanning a pseudo-connection:
            "double" walks are not allowed.)
        connection_arrival_time: float, int, optional
            used for computing the walking label if dep_time, i.e., connection.arrival_stop_next_departure_time, is infinity)
        connection: connection object

        Returns
        -------
        pareto_optimal_labels : set
            Set of Labels
        """
        walk_labels = list()
        # walk label towards target
        if first_leg_can_be_walk and self._walk_to_target_duration != float('inf'):
            # add walk_label
            if connection_arrival_time is not None:
                walk_labels.append(self._get_label_to_target(connection_arrival_time))
            else:
                walk_labels.append(self._get_label_to_target(dep_time))

        # if dep time is larger than the largest dep time -> only walk labels are possible
        if dep_time in self.dep_times_to_index:
            assert (dep_time != float('inf'))
            index = self.dep_times_to_index[dep_time]
            labels = self._label_bags[index]
            pareto_optimal_labels = merge_pareto_frontiers(labels, walk_labels)
        else:
            pareto_optimal_labels = walk_labels

        if not first_leg_can_be_walk:
            pareto_optimal_labels = [label for label in pareto_optimal_labels if not label.first_leg_is_walk]
        return pareto_optimal_labels

    def _get_label_to_target(self, departure_time):
        if departure_time != float('inf') and self._walk_to_target_duration != float('inf'):
            if self._walk_to_target_duration == 0:
                first_leg_is_walk = False
            else:
                first_leg_is_walk = True
            if self.label_class == LabelTimeBoardingsAndRoute or self.label_class == LabelTimeAndRoute:
                if self._walk_to_target_duration > 0:
                    walk_connection = Connection(self.node_id,
                                                 self.closest_target,
                                                 departure_time,
                                                 departure_time + self._walk_to_target_duration,
                                                 Connection.WALK_TRIP_ID,
                                                 Connection.WALK_SEQ,
                                                 is_walk=True
                                                 )
                else:
                    walk_connection = None
                if self.label_class == LabelTimeAndRoute:
                    label = self.label_class(departure_time=float(departure_time),
                                             arrival_time_target=float(departure_time + self._walk_to_target_duration),
                                             movement_duration=self._walk_to_target_duration,
                                             first_leg_is_walk=first_leg_is_walk,
                                             connection=walk_connection)
                else:
                    label = self.label_class(departure_time=float(departure_time),
                                             arrival_time_target=float(departure_time + self._walk_to_target_duration),
                                             movement_duration=self._walk_to_target_duration,
                                             n_boardings=0,
                                             first_leg_is_walk=first_leg_is_walk,
                                             connection=walk_connection)
            else:
                label = self.label_class(departure_time=float(departure_time),
                                         arrival_time_target=float(departure_time + self._walk_to_target_duration),
                                         n_boardings=0,
                                         first_leg_is_walk=first_leg_is_walk)

            return label
        else:
            return None

    def get_labels_for_real_connections(self):
        self._closed = True
        if self._real_connection_labels is None:
            self._compute_real_connection_labels()
        return self._real_connection_labels

    def get_final_optimal_labels(self):
        """
        Get pareto-optimal labels.

        Returns
        -------
        """
        assert self._finalized, "finalize() first!"
        return self._final_pareto_optimal_labels

    def finalize(self, neighbor_label_bags=None, walk_durations=None, departure_arrival_stop_pairs=None):
        """
        Parameters
        ----------
        neighbor_label_bags: list
            each list element is a list of labels corresponding to a neighboring node
             (note: only labels with first connection being a departure should be included)
        walk_durations: list
        departure_arrival_stop_pairs: list of tuples
        Returns
        -------
        None
        """
        assert (not self._finalized)
        if self._final_pareto_optimal_labels is None:
            self._compute_real_connection_labels()
        if neighbor_label_bags is not None:
            assert (len(walk_durations) == len(neighbor_label_bags))
            self._compute_final_pareto_optimal_labels(neighbor_label_bags,
                                                      walk_durations,
                                                      departure_arrival_stop_pairs)
        else:
            self._final_pareto_optimal_labels = self._real_connection_labels
        self._finalized = True
        self._closed = True

    def _compute_real_connection_labels(self):
        pareto_optimal_labels = []
        # do not take those bags with first event is a pseudo-connection
        for dep_time in self._connection_dep_times:
            index = self.dep_times_to_index[dep_time]
            pareto_optimal_labels.extend([label for label in self._label_bags[index] if not label.first_leg_is_walk])
        if self.label_class == LabelTimeWithBoardingsCount or self.label_class == LabelTime \
                or self.label_class == LabelTimeBoardingsAndRoute:
            pareto_optimal_labels = [label for label in pareto_optimal_labels
                                     if label.duration() < self._walk_to_target_duration]

        if self.label_class == LabelVehLegCount and self._walk_to_target_duration < float('inf'):
            pareto_optimal_labels.append(LabelVehLegCount(0))
        self._real_connection_labels = [label.get_copy() for label in compute_pareto_front(pareto_optimal_labels,
                                                                                           finalization=True)]

    def _compute_final_pareto_optimal_labels(self, neighbor_label_bags, walk_durations, departure_arrival_stops):
        labels_from_neighbors = []
        for i, (label_bag, walk_duration)in enumerate(zip(neighbor_label_bags, walk_durations)):
            for label in label_bag:
                if self.label_class == LabelTimeBoardingsAndRoute or self.label_class == LabelTimeAndRoute:
                    departure_arrival_tuple = departure_arrival_stops[i]
                    departure_time = label.departure_time - walk_duration
                    arrival_time = label.departure_time
                    connection = Connection(departure_arrival_tuple[0],
                                            departure_arrival_tuple[1],
                                            departure_time,
                                            arrival_time,
                                            Connection.WALK_TRIP_ID,
                                            Connection.WALK_SEQ,
                                            is_walk=True)
                    labels_from_neighbors.append(label.get_copy_with_walk_added(walk_duration, connection))
                else:
                    labels_from_neighbors.append(label.get_copy_with_walk_added(walk_duration))

        pareto_front = compute_pareto_front(self._real_connection_labels +
                                            labels_from_neighbors,
                                            finalization=True)

        if pareto_front and hasattr(pareto_front[0], "duration"):
            self._final_pareto_optimal_labels = list(filter(lambda label: label.duration() < self._walk_to_target_duration, pareto_front))
        else:
            self._final_pareto_optimal_labels = pareto_front





import pandas as pd

from six import string_types
from gtfspy.warnings_container import WarningsContainer
from gtfspy.gtfs import GTFS
from gtfspy.util import str_time_to_day_seconds
from gtfspy.util import source_csv_to_pandas

"""
Input: sourcefile(s) of the gtfs data, created sqlite database
output: error messages for failed checks
- Assert that the sqlite database exists and that it is not corrupted
    -> it is possible to create a gtfs object
    -
- Row count of each table, check that each rowcount matches the imported files
    - Open each table in sqlite database and all text files in the zips
    - Compare
    - Should GTFS source file dirs be added to metadata so that the checkup can be made later?

"""


WARNING_DANGLING_STOPS_VS_STOP_TIMES_AND_PARENT_STOPS = "stops are not referenced in stop_times and are not marked as a parent_stop (it is ok/possible that there are some)"
WARNING_DANGLING_STOP_TIMES_VS_STOPS = "stop_times referencing to missing stop"
WARNING_DANGLING_STOP_TIMES_VS_TRIPS = "stop_times not referenced in trips found"
WARNING_DANGLING_TRIPS_VS_STOP_TIMES = "trips with missing stop_times found"
WARNING_DANGLING_TRIPS_VS_DAYS = "trips not referenced in days found (this is possible due to some combinations of calendar and calendar_dates)"
WARNING_DANGLING_TRIPS_VS_CALENDAR = "trips whose service_I not referenced in calendar found"
WARNING_DANGLING_TRIPS_VS_ROUTES = "trips not referenced in routes found"
WARNING_DANGLING_DAYS_VS_TRIPS = "days not referenced in trips found"
WARNING_DANGLING_ROUTES_VS_TRIPS = "routes not referenced in trips found"


DANGLER_QUERIES = [
    'SELECT count(*) FROM stops '
        'LEFT JOIN stop_times ON(stop_times.stop_I=stops.stop_I) '
        'LEFT JOIN stops as parents ON(stops.stop_I=parents.parent_I) '
        'WHERE (stop_times.stop_I IS NULL AND parents.parent_I IS NULL)',
    'SELECT count(*) FROM stop_times LEFT JOIN stops ON(stop_times.stop_I=stops.stop_I) WHERE stops.stop_I IS NULL',
    'SELECT count(*) FROM stop_times LEFT JOIN trips ON(stop_times.trip_I=trips.trip_I) WHERE trips.trip_I IS NULL',
    'SELECT count(*) FROM trips LEFT JOIN stop_times ON(stop_times.trip_I=trips.trip_I) WHERE stop_times.trip_I IS NULL',
    'SELECT count(*) FROM trips LEFT JOIN days ON(days.trip_I=trips.trip_I) WHERE days.trip_I IS NULL',
    'SELECT count(*) FROM trips LEFT JOIN calendar ON(calendar.service_I=trips.service_I) WHERE trips.service_I IS NULL',
    'SELECT count(*) FROM trips LEFT JOIN routes ON(routes.route_I=trips.route_I) WHERE routes.route_I IS NULL',
    'SELECT count(*) FROM days LEFT JOIN trips ON(days.trip_I=trips.trip_I) WHERE trips.trip_I IS NULL',
    'SELECT count(*) FROM routes LEFT JOIN trips ON(routes.route_I=trips.route_I) WHERE trips.route_I IS NULL'
]

DANGLER_WARNINGS = [
    WARNING_DANGLING_STOPS_VS_STOP_TIMES_AND_PARENT_STOPS,
    WARNING_DANGLING_STOP_TIMES_VS_STOPS,
    WARNING_DANGLING_STOP_TIMES_VS_TRIPS,
    WARNING_DANGLING_TRIPS_VS_STOP_TIMES,
    WARNING_DANGLING_TRIPS_VS_DAYS,
    WARNING_DANGLING_TRIPS_VS_CALENDAR,
    WARNING_DANGLING_TRIPS_VS_ROUTES,
    WARNING_DANGLING_DAYS_VS_TRIPS,
    WARNING_DANGLING_ROUTES_VS_TRIPS,
]

DB_TABLE_NAME_TO_FIELDS_WHERE_NULL_NOT_OK = {
    'agencies': ['agency_I', 'agency_id', "timezone"],
    'stops': ['stop_I', 'stop_id', 'lat', 'lon'],
    'routes': ['route_I', 'route_id', 'type'],
    'trips': ['trip_I', 'trip_id', 'service_I', "route_I"],
    'stop_times': ["trip_I", "stop_I", "arr_time_ds", "dep_time_ds"],
    'calendar': ['service_id', 'service_I', 'm', "t", "w", "th", "f", "s", "su", "start_date", "end_date"],
    'calendar_dates': ['service_I', 'date', 'exception_type'],
    'days': ["date","day_start_ut","trip_I"],
    'shapes': ["shape_id", "lat", "lon", "seq"],
    'stop_distances': ["from_stop_I", "to_stop_I", "d"]
}

DB_TABLE_NAME_TO_FIELDS_WHERE_NULL_OK_BUT_WARN = {
    'agencies': ['name', "url"],
    'stops': ['name'],
    'routes': ['name', 'long_name'],
    'trips': [],
    'calendar': [],
    'calendar_dates': [],
    'days': [],
    'shapes': [],
    'stop_times': [],
    'stop_distances': ["d_walk"]
}

DB_TABLE_NAMES = list(sorted(DB_TABLE_NAME_TO_FIELDS_WHERE_NULL_NOT_OK.keys()))

DB_TABLE_NAME_TO_SOURCE_FILE = {
    'agencies': "agency",
    'routes': "routes",
    'trips': "trips",
    'calendar': "calendar",
    'calendar_dates': "calendar_dates",
    'stop_times': "stop_times",
    'stops': "stops",
    "shapes": 'shapes'
}

DB_TABLE_NAME_TO_ROWS_MISSING_WARNING = {}
for _db_table_name in DB_TABLE_NAMES:
    DB_TABLE_NAME_TO_ROWS_MISSING_WARNING[_db_table_name] = "Rows missing in {table}".format(table=_db_table_name)
DB_TABLE_NAME_TO_ROWS_MISSING_WARNING["calendar"] = "There are extra/missing rows in calendar that cannot be explained " \
                                                    "by dummy entries required by the calendar_dates table."

for dictionary in [DB_TABLE_NAME_TO_SOURCE_FILE, DB_TABLE_NAME_TO_ROWS_MISSING_WARNING]:
    for key in dictionary.keys():
        assert key in DB_TABLE_NAMES

for key in DB_TABLE_NAME_TO_FIELDS_WHERE_NULL_OK_BUT_WARN.keys():
    assert key in DB_TABLE_NAME_TO_FIELDS_WHERE_NULL_NOT_OK

#SOURCE_TABLE_NAMES = ['agency', 'routes', 'trips', 'calendar', 'calendar_dates', 'stop_times', 'stops', 'shapes']


class ImportValidator(object):

    def __init__(self, gtfssource, gtfs, verbose=True):
        """
        Parameters
        ----------
        gtfs_sources: list, string, dict
            list of paths to the strings, or a dictionary directly containing the gtfs data directly
        gtfs: gtfspy.gtfs.GTFS, or path to a relevant .sqlite GTFS database
        verbose: bool
            Whether or not to print warnings on-the-fly.
        """
        if isinstance(gtfssource, string_types + (dict,)):
            self.gtfs_sources = [gtfssource]
        else:
            assert isinstance(gtfssource, list)
            self.gtfs_sources = gtfssource
        assert len(self.gtfs_sources) > 0, "There needs to be some source files for validating an import"

        if not isinstance(gtfs, GTFS):
            self.gtfs = GTFS(gtfs)
        else:
            self.gtfs = gtfs

        self.location = self.gtfs.get_location_name()
        self.warnings_container = WarningsContainer()
        self.verbose=verbose

    def validate_and_get_warnings(self):
        self.warnings_container.clear()
        self._validate_table_row_counts()
        self._validate_no_null_values()
        self._validate_danglers()
        return self.warnings_container

    def _validate_table_row_counts(self):
        """
        Imports source .txt files, checks row counts and then compares the rowcounts with the gtfsobject
        :return:
        """
        for db_table_name in DB_TABLE_NAME_TO_SOURCE_FILE.keys():
            table_name_source_file = DB_TABLE_NAME_TO_SOURCE_FILE[db_table_name]
            row_warning_str = DB_TABLE_NAME_TO_ROWS_MISSING_WARNING[db_table_name]

            # Row count in GTFS object:
            database_row_count = self.gtfs.get_row_count(db_table_name)

            # Row counts in source files:
            source_row_count = 0
            for gtfs_source in self.gtfs_sources:
                frequencies_in_source = source_csv_to_pandas(gtfs_source, 'frequencies.txt')
                try:
                    if table_name_source_file == 'trips' and not frequencies_in_source.empty:
                        source_row_count += self._frequency_generated_trips_rows(gtfs_source)

                    elif table_name_source_file == 'stop_times' and not frequencies_in_source.empty:
                        source_row_count += self._compute_number_of_frequency_generated_stop_times(gtfs_source)
                    else:
                        df = source_csv_to_pandas(gtfs_source, table_name_source_file)

                        source_row_count += len(df.index)
                except IOError as e:
                    if hasattr(e, "filename") and db_table_name in e.filename:
                        pass
                    else:
                        raise e


            if source_row_count == database_row_count and self.verbose:
                print("Row counts match for " + table_name_source_file + " between the source and database ("
                      + str(database_row_count) + ")")
            else:
                difference = database_row_count - source_row_count
                ('Row counts do not match for ' + str(table_name_source_file) + ': (source=' + str(source_row_count) +
                      ', database=' + str(database_row_count) + ")")
                if table_name_source_file == "calendar" and difference > 0:
                    query = "SELECT count(*) FROM (SELECT * FROM calendar ORDER BY service_I DESC LIMIT " \
                            + str(int(difference)) + \
                            ") WHERE start_date=end_date AND m=0 AND t=0 AND w=0 AND th=0 AND f=0 AND s=0 AND su=0"
                    number_of_entries_added_by_calendar_dates_loader = self.gtfs.execute_custom_query(query).fetchone()[
                        0]
                    if number_of_entries_added_by_calendar_dates_loader == difference and self.verbose:
                        print("    But don't worry, the extra entries seem to just dummy entries due to calendar_dates")
                    else:
                        if self.verbose:
                            print("    Reason for this is unknown.")
                        self.warnings_container.add_warning(row_warning_str, self.location, difference)
                else:
                    self.warnings_container.add_warning(row_warning_str, self.location, difference)


    def _validate_no_null_values(self):
        """
        Loads the tables from the gtfs object and counts the number of rows that have null values in
        fields that should not be null. Stores the number of null rows in warnings_container
        """
        for table in DB_TABLE_NAMES:
            null_not_ok_warning = "Null values in must-have columns in table {table}".format(table=table)
            null_warn_warning = "Null values in good-to-have columns in table {table}".format(table=table)
            null_not_ok_fields = DB_TABLE_NAME_TO_FIELDS_WHERE_NULL_NOT_OK[table]
            null_warn_fields = DB_TABLE_NAME_TO_FIELDS_WHERE_NULL_OK_BUT_WARN[table]

            # CW, TODO: make this validation source by source
            df = self.gtfs.get_table(table)

            for warning, fields in zip([null_not_ok_warning, null_warn_warning], [null_not_ok_fields, null_warn_fields]):
                null_unwanted_df = df[fields]
                rows_having_null = null_unwanted_df.isnull().any(1)
                if sum(rows_having_null) > 0:
                    rows_having_unwanted_null = df[rows_having_null.values]
                    self.warnings_container.add_warning(warning, rows_having_unwanted_null, len(rows_having_unwanted_null))


    def _validate_danglers(self):
        """
        Checks for rows that are not referenced in the the tables that should be linked

        stops <> stop_times using stop_I
        stop_times <> trips <> days, using trip_I
        trips <> routes, using route_I
        :return:
        """
        for query, warning in zip(DANGLER_QUERIES, DANGLER_WARNINGS):
            dangler_count = self.gtfs.execute_custom_query(query).fetchone()[0]
            if dangler_count > 0:
                if self.verbose:
                    print(str(dangler_count) + " " + warning)
                self.warnings_container.add_warning(warning, self.location, count=dangler_count)

    def _frequency_generated_trips_rows(self, gtfs_soure_path, return_df_freq=False):
        """
        This function calculates the equivalent rowcounts for trips when
        taking into account the generated rows in the gtfs object
        Parameters
        ----------
        gtfs_soure_path: path to the source file
        param txt: txt file in question
        :return: sum of all trips
        """
        df_freq = source_csv_to_pandas(gtfs_soure_path, 'frequencies')
        df_trips = source_csv_to_pandas(gtfs_soure_path, "trips")
        df_freq['n_trips'] = df_freq.apply(lambda row: len(range(str_time_to_day_seconds(row['start_time']),
                                                                 str_time_to_day_seconds(row['end_time']),
                                                                 row['headway_secs'])), axis=1)
        df_trips_freq = pd.merge(df_freq, df_trips, how='outer', on='trip_id')
        n_freq_generated_trips = int(df_trips_freq['n_trips'].fillna(1).sum(axis=0))
        if return_df_freq:
            return df_trips_freq
        else:
            return n_freq_generated_trips

    def _compute_number_of_frequency_generated_stop_times(self, gtfs_source_path):
        """
        Parameters
        ----------
        Same as for "_frequency_generated_trips_rows" but for stop times table
        gtfs_source_path:
        table_name:

        Return
        ------
        """
        df_freq = self._frequency_generated_trips_rows(gtfs_source_path, return_df_freq=True)
        df_stop_times = source_csv_to_pandas(gtfs_source_path, "stop_times")
        df_stop_freq = pd.merge(df_freq, df_stop_times, how='outer', on='trip_id')
        return int(df_stop_freq['n_trips'].fillna(1).sum(axis=0))
